import json
import logging
import sys
from copy import copy
from importlib import metadata
from itertools import pairwise
from math import sqrt
from os import cpu_count
from pathlib import Path
from typing import Iterable, Literal, Sequence

import cv2
import h5py
import numpy as np

from .utils import (
    CustomError,
    Episode,
    Timer,
    assert_all_files_exist,
    assert_knowledge_transfer_is_possible,
    build_ROI_mask_from_list,
    create_dir,
    json_default,
    json_object_hook,
    remove_dir,
    remove_file,
    resolve_path,
    track,
)


class Video:
    """
    A class containing the main features of the video.

    This class includes properties of the video by itself, user defined
    parameters for the tracking, and other properties that are generated
    throughout the tracking process.

    We use this class as a storage of data coming from different processes.
    However, this is bad practice and it will change in the future.
    """

    velocity_threshold: float
    erosion_kernel_size: int
    ratio_accumulated_images: float
    accumulation_folder: Path
    # FIXME it should depend on self.session_folder
    # return self.session_folder / f"accumulation_{self.accumulation_trial}"
    individual_fragments_stats: dict
    percentage_of_accumulated_images: list[float]
    # TODO: move to accumulation_manager.py
    # TODO: move to accumulation_manager.py
    session_folder: Path
    # TODO remove these defaults, they are already in __main__
    setup_points: dict[str, list[tuple[int, int]]]

    median_body_length: float
    """median of the diagonals of individual blob's bounding boxes"""

    # TODO: move tracker.py
    first_frame_first_global_fragment: list

    # During validation (in validation GUI)
    identities_groups: dict
    """Groups of identities stored during the validation of the tracking
    in the validation GUI. This is useful to group identities in different
    classes depending on the experiment.

    This feature was coded because some users require indicating classes
    of individuals but we do not use it in the lab."""
    episodes: list[Episode]
    """Indicates the starting and ending frames of each video episode.
    Video episodes are used for parallelization of some processes"""

    original_width: int
    """Original video width in pixels. It does not consider the resolution
    reduction factor defined by the user"""
    original_height: int
    """Original video width in pixels. It does not consider the resolution
    reduction factor defined by the user"""
    frames_per_second: int
    """Video frame rate in frames per second obtained by OpenCV from the
    video file"""
    accumulation_statistics: dict[str, list]
    accumulation_statistics_data: list[dict[str, list]]
    number_of_error_frames: int = -1
    """The number of frames with more blobs than animals. Set on animals_detection."""
    estimated_accuracy: float | None = None
    accumulation_trial: int = 0
    identities_labels: list[str] | None = None
    """A list with a name for every identity. Defined and used in validator"""
    background_from_segmentation_gui: np.ndarray | None = None
    """Background set by segmentation app to save when the app closes"""

    video_paths: list[Path] = []
    """List of paths to the different files the video is composed of.
    If the video is a single file, the list will have length 1"""
    number_of_animals: int = 0
    intensity_ths: None | Sequence[int] = None
    area_ths: None | Sequence[int] = None
    # bkg_model: None | np.ndarray = None
    session: str = "no_name"
    output_dir: Path | None | str = None
    tracking_intervals: list | None = None
    resolution_reduction: float = 1.0
    roi_list: list[str] | str | None = None
    use_bkg: bool = False
    knowledge_transfer_folder: None | Path = None
    check_segmentation: bool = False
    identity_transfer: bool = False
    track_wo_identities: bool = False
    frames_per_episode: int = 500
    background_subtraction_stat: Literal["median", "mean", "max", "min"] = "median"
    number_of_frames_for_background: int = 50
    number_of_parallel_workers: int = 0
    data_policy: Literal[
        "trajectories", "validation", "knowledge_transfer", "idmatcher.ai", "all"
    ] = "all"
    id_image_size: list[int] = []
    """ Shape of the Blob's identification images (width, height, n_channels)"""
    protocol3_action: Literal["ask", "abort", "continue"] = "ask"
    convert_trajectories_to_csv_and_json: bool = True
    add_time_column_to_csv: bool = False
    """Add a time column (in seconds) to csv trajectory filesy"""
    version: str
    """Version of idtracker.ai"""

    def set_parameters(self, **parameters):
        """Sets parameters to self only if they are present in the class annotations.
        The set of non recognized parameters names is returned"""
        non_recognized_parameters: set[str] = set()
        for param, value in parameters.items():
            lower_param = param.lower()
            if lower_param in self.__class__.__annotations__:
                setattr(self, lower_param, value)
            else:
                non_recognized_parameters.add(param)
        return non_recognized_parameters

    def prepare_tracking(self):
        """Initializes the video object, checking all parameters"""
        logging.debug("Initializing Video object")
        self.version = metadata.version("idtrackerai")

        if not isinstance(self.video_paths, list):
            video_paths = [self.video_paths]
        else:
            video_paths = self.video_paths
        self.assert_video_paths(video_paths)
        self.video_paths = [Path(path).expanduser().resolve() for path in video_paths]
        logging.info(
            "Setting video_paths to:\n    " + "\n    ".join(map(str, self.video_paths))
        )

        if self.area_ths is None:
            raise CustomError("Missing area thresholds parameter")

        if self.intensity_ths is None:
            raise CustomError("Missing intensity thresholds parameter")

        self.accumulation_statistics_data = []

        if self.knowledge_transfer_folder is not None:
            self.knowledge_transfer_folder = resolve_path(
                self.knowledge_transfer_folder
            )
            if not self.knowledge_transfer_folder.exists():
                raise CustomError(
                    f'Knowledge transfer folder "{self.knowledge_transfer_folder}" not'
                    " found"
                )

        self.original_width, self.original_height, self.frames_per_second = (
            self.get_info_from_video_paths(self.video_paths)
        )
        self.number_of_frames, _, self.tracking_intervals, self.episodes = (
            self.get_processing_episodes(
                self.video_paths, self.frames_per_episode, self.tracking_intervals
            )
        )

        logging.info(
            f"The video has {self.number_of_frames} "
            f"frames ({self.number_of_episodes} episodes)"
        )
        if len(self.episodes) < 10:
            for e in self.episodes:
                video_name = self.video_paths[e.video_path_index].name
                logging.info(
                    f"\tEpisode {e.index}, frames ({e.local_start} "
                    f"=> {e.local_end}) of /{video_name}"
                )
        assert self.number_of_episodes > 0

        if self.output_dir is not None:
            self.session_folder = (
                resolve_path(self.output_dir) / f"session_{self.session.strip()}"
            )
        else:
            self.session_folder = (
                self.video_folder / f"session_{self.session.strip()}"
            ).resolve()
        create_dir(self.session_folder)
        create_dir(self.preprocessing_folder)

        self.ROI_mask = build_ROI_mask_from_list(
            self.roi_list,
            self.resolution_reduction,
            self.original_width,
            self.original_height,
        )

        if isinstance(self.id_image_size, int):
            self.id_image_size = [self.id_image_size, self.id_image_size, 1]
        else:
            self.id_image_size = []

        if self.knowledge_transfer_folder is not None:
            self.id_image_size = assert_knowledge_transfer_is_possible(
                self.knowledge_transfer_folder, self.n_animals
            )

        if self.number_of_parallel_workers <= 0:
            computer_CPUs = cpu_count()
            if computer_CPUs is not None:
                if self.number_of_parallel_workers == 0:
                    self.number_of_parallel_workers = (computer_CPUs + 1) // 2
                elif self.number_of_parallel_workers < 0:
                    self.number_of_parallel_workers += computer_CPUs
        logging.info("Number of parallel jobs: %d", self.number_of_parallel_workers)

        if self.number_of_animals == 0 and not self.track_wo_identities:
            raise CustomError(
                "Cannot track with an undefined number of animals (n_animals = 0)"
                " when tracking with identities"
            )

        self.bkg_model = self.background_from_segmentation_gui  # has a setter
        self.__dict__.pop("background_from_segmentation_gui", None)

        self.first_frame_first_global_fragment = []
        self.identities_groups = {}
        self.setup_points = {}

        # Processes timers
        self.general_timer = Timer("Tracking session")
        self.detect_animals_timer = Timer("Animal detection")
        self.crossing_detector_timer = Timer("Crossing detection")
        self.fragmentation_timer = Timer("Fragmentation")
        self.tracking_timer = Timer("Tracking")
        self.protocol1_timer = Timer("Protocol 1")
        self.protocol2_timer = Timer("Protocol 2")
        self.protocol3_pretraining_timer = Timer("Protocol 3 pre-training")
        self.protocol3_accumulation_timer = Timer("Protocol 3 accumulation")
        self.identify_timer = Timer("Identification")
        self.impossible_jumps_timer = Timer("Impossible jumps correction")
        self.crossing_solver_timer = Timer("Crossings solver")
        self.create_trajectories_timer = Timer("Trajectories creation")

        self.general_timer.start()

    def __str__(self) -> str:
        return f"<session {self.session_folder}>"

    def set_id_image_size(self, median_body_length: int | float, reset=False):
        self.median_body_length = median_body_length
        if reset or not self.id_image_size:
            side_length = int(median_body_length / sqrt(2))
            side_length += side_length % 2
            self.id_image_size = [side_length, side_length, 1]
        logging.info(f"Identification image size set to {self.id_image_size}")

    @property
    def n_animals(self):
        return self.number_of_animals

    @property
    def single_animal(self) -> bool:
        return self.n_animals == 1

    @property
    def bkg_model(self) -> np.ndarray | None:
        if self.background_path.is_file():
            return cv2.imread(str(self.background_path))[..., 0]
        return None

    @bkg_model.setter
    def bkg_model(self, bkg: np.ndarray | None):
        if bkg is None:
            del self.bkg_model
        else:
            cv2.imwrite(str(self.background_path), bkg)
            logging.info(f"Background saved at {self.background_path}")

    @bkg_model.deleter
    def bkg_model(self):
        self.background_path.unlink(missing_ok=True)

    @property
    def ROI_list(self):
        """Fixes compatibility issues"""
        return self.roi_list

    @property
    def ROI_mask(self) -> np.ndarray | None:
        if self.ROI_mask_path.is_file():
            return cv2.imread(str(self.ROI_mask_path))[..., 0]
        return None

    @ROI_mask.setter
    def ROI_mask(self, mask: np.ndarray | None):
        if mask is None:
            del self.ROI_mask
        else:
            cv2.imwrite(str(self.ROI_mask_path), mask)
            logging.info(f"ROI mask saved at {self.ROI_mask_path}")

    @ROI_mask.deleter
    def ROI_mask(self):
        self.ROI_mask_path.unlink(missing_ok=True)

    @property
    def video_folder(self) -> Path:
        """Directory where video was stored. Parent of video_path.

        Returns
        -------
        str
            Path to the video folder where the video to be tracked was stored.
        """
        return self.video_paths[0].parent

    @property
    def number_of_episodes(self):
        """Number of episodes in which the video is splitted for parallel
        processing.

        Returns
        -------
        int
            Number of parts in which the videos is splitted.

        See Also
        --------
        :int:`~idtrackerai.constants.FRAMES_PER_EPISODE`
        """
        return len(self.episodes)

    @property
    def width(self):
        """Video width in pixels after applying the resolution reduction
        factor.

        Returns
        -------
        int
            Video width in pixels after applying the resolution reduction
            factor defined by the user.
        """
        return int(self.original_width * self.resolution_reduction + 0.5)

    @property
    def height(self):
        """Video height in pixels after applying the resolution reduction
        factor.

        Returns
        -------
        int
            Video height in pixels after applying the resolution reduction
            factor.
        """
        return int(self.original_height * self.resolution_reduction + 0.5)

    # TODO: move to crossings_detection.py
    @property
    def median_body_length_full_resolution(self):
        """Median body length in pixels in full frame resolution
        (i.e. without considering the resolution reduction factor)
        """
        return self.median_body_length / self.resolution_reduction

    # Paths and folders
    # TODO: The different processes should create and store the path to the
    # folder where they save the data
    @property
    def preprocessing_folder(self) -> Path:
        return self.session_folder / "preprocessing"

    @property
    def background_path(self) -> Path:
        return self.preprocessing_folder / "background.png"

    @property
    def ROI_mask_path(self) -> Path:
        return self.preprocessing_folder / "ROI_mask.png"

    @property
    def trajectories_folder(self) -> Path:
        return self.session_folder / "trajectories"

    @property
    def crossings_detector_folder(self) -> Path:
        return self.session_folder / "crossings_detector"

    @property
    def pretraining_folder(self) -> Path:
        return self.session_folder / "pretraining"

    @property
    def individual_videos_folder(self) -> Path:
        return self.session_folder / "individual_videos"

    @property
    def auto_accumulation_folder(self) -> Path:
        return self.session_folder / f"accumulation_{self.accumulation_trial}"

    @property
    def id_images_folder(self) -> Path:
        return self.session_folder / "identification_images"

    # TODO: This should probably be the only path that should be stored in
    # Video.

    @property
    def blobs_path(self) -> Path:
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs.pickle"

    @property
    def blobs_no_gaps_path(self) -> Path:
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs_no_gaps.pickle"

    @property
    def blobs_path_validated(self) -> Path:
        return self.preprocessing_folder / "list_of_blobs_validated.pickle"

    @property
    def idmatcher_results_path(self) -> Path:
        return self.session_folder / "matching_results"

    @property
    def global_fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_global_fragments.json"

    @property
    def fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_fragments.json"

    @property
    def path_to_video_object(self) -> Path:
        return self.session_folder / "video_object.json"

    @property
    def ground_truth_path(self) -> Path:
        return self.video_folder / "_groundtruth.npy"

    @property
    def segmentation_data_folder(self) -> Path:
        return self.session_folder / "segmentation_data"

    @property
    def id_images_file_paths(self) -> list[Path]:
        return [
            self.id_images_folder / f"id_images_{e}.hdf5"
            for e in range(self.number_of_episodes)
        ]

    # Methods
    def save(self):
        """Saves the instantiated Video object"""
        logging.info(f"Saving video object in {self.path_to_video_object}")
        dict_to_save = copy(self.__dict__)
        dict_to_save.pop("episodes", None)
        self.path_to_video_object.write_text(
            json.dumps(dict_to_save, default=json_default, indent=4)
        )

    @classmethod
    def load(cls, path: Path | str, video_paths_dir: Path | None = None) -> "Video":
        """Load a video object stored in a JSON file"""
        path = resolve_path(path)
        logging.info(f"Loading Video from {path}")
        if not path.is_file():
            path /= "video_object.json"
            if not path.is_file():
                path = path.with_suffix(".npy")
                if not path.is_file():
                    raise FileNotFoundError(f"{path} not found")

        if path.suffix == ".npy":
            video_dict = cls.open_from_v4(path)
        else:
            with open(path, "r", encoding="utf_8") as file:
                video_dict = json.load(file, object_hook=json_object_hook)

        if "n_animals" not in video_dict and "number_of_animals" in video_dict:
            video_dict["n_animals"] = video_dict["number_of_animals"]

        video = cls.__new__(cls)
        video.__dict__.update(video_dict)
        video.update_paths(path.parent, video_paths_dir)
        try:
            _, _, _, video.episodes = video.get_processing_episodes(
                video.video_paths, video.frames_per_episode, video.tracking_intervals
            )
        except AttributeError:
            logging.warning(
                "Could not load video episodes probably due to loading an old version"
                " session"
            )
        return video

    @classmethod
    def open_from_v4(cls, path: Path) -> dict:
        from . import network

        logging.warning("Loading from v4: %s", path)

        # v4 compatibility
        sys.modules["idtrackerai.tracker.network.network_params"] = network
        _dict: dict = np.load(path, allow_pickle=True).item().__dict__
        del sys.modules["idtrackerai.tracker.network.network_params"]

        _dict["version"] = "4.0.12 or below"
        _dict["video_paths"] = list(map(Path, _dict.pop("_video_paths")))
        _dict["session_folder"] = path.parent
        _dict["median_body_length"] = _dict.pop("_median_body_length")
        _dict["frames_per_second"] = _dict.pop("_frames_per_second")
        _dict["original_width"] = _dict.pop("_original_width")
        _dict["original_height"] = _dict.pop("_original_height")
        _dict["number_of_frames"] = _dict.pop("_number_of_frames")
        _dict["identities_groups"] = _dict.pop("_identities_groups")
        _dict["id_image_size"] = list(_dict.pop("_identification_image_size"))
        _dict["setup_points"] = _dict.pop("_setup_points")
        _dict["number_of_animals"] = _dict["_user_defined_parameters"][
            "number_of_animals"
        ]
        _dict["tracking_intervals"] = _dict["_user_defined_parameters"][
            "tracking_interval"
        ]
        _dict["resolution_reduction"] = _dict["_user_defined_parameters"][
            "resolution_reduction"
        ]
        _dict["track_wo_identities"] = _dict["_user_defined_parameters"][
            "track_wo_identification"
        ]
        _dict["accumulation_folder"] = (
            path.parent / Path(_dict.pop("_accumulation_folder")).name
        )
        _dict["_user_defined_parameters"].pop("mask")
        return _dict

    def update_paths(
        self, new_video_object_path: Path, user_video_paths_dir: Path | None = None
    ):
        """Update paths of objects (e.g. blobs_path, preprocessing_folder...)
        according to the new location of the new video object given
        by `new_video_object_path`.

        Parameters
        ----------
        new_video_object_path : str
            Path to a video_object.npy
        """
        logging.info(
            f"Searching video files: {[str(path.name) for path in self.video_paths]}"
        )
        folder_candidates: set[Path | None] = {
            user_video_paths_dir,
            self.video_paths[0],
            new_video_object_path,
            new_video_object_path.parent,
            self.session_folder.parent,
            self.session_folder,
        }

        for folder_candidate in folder_candidates:
            if folder_candidate is None:
                continue
            if folder_candidate.is_file():
                folder_candidate = folder_candidate.parent

            candidate_new_video_paths = [
                folder_candidate / path.name for path in self.video_paths
            ]

            try:
                assert_all_files_exist(candidate_new_video_paths)
            except FileNotFoundError:
                continue

            logging.info(f"All video files found on {folder_candidate}")
            found = True
            break
        else:
            found = False
            candidate_new_video_paths = []
            logging.error(f"Video file paths not found: {self.video_paths}")

        need_to_save = False
        if self.session_folder != new_video_object_path:
            logging.info(
                f"Updated session folder from {self.session_folder} to"
                f" {new_video_object_path}"
            )
            self.session_folder = new_video_object_path
            need_to_save = True

        if found and self.video_paths != candidate_new_video_paths:
            logging.info("Updating new video files paths")
            self.video_paths = candidate_new_video_paths
            need_to_save = True

        if need_to_save:
            self.save()

    @staticmethod
    def assert_video_paths(video_paths: Iterable[Path | str]):
        if not video_paths:
            raise CustomError("Empty Video paths list")

        for path in video_paths:
            path = Path(path).expanduser().resolve()
            if not path.is_file():
                raise CustomError(f'Video file "{path}" not found')

            readable = cv2.VideoCapture(str(path)).grab()
            if not readable:
                raise CustomError(f'Video file "{path}" not readable by OpenCV.')

    @staticmethod
    def get_info_from_video_paths(video_paths: Iterable[Path | str]):
        """Gets some information about the video from the video file itself.

        Returns:
            width: int, height: int, fps: int
        """

        widths, heights, fps = [], [], []
        for path in video_paths:
            cap = cv2.VideoCapture(str(path))
            widths.append(int(cap.get(3)))
            heights.append(int(cap.get(4)))

            try:
                fps.append(int(cap.get(5)))
            except cv2.error:
                logging.warning(f"Cannot read frame per second for {path}")
                fps.append(None)
            cap.release()

        if len(set(widths)) != 1 or len(set(heights)) != 1:
            raise CustomError("Video paths have different resolutions")

        if len(set(fps)) != 1:
            fps = [int(np.mean(fps))]
            logging.warning(
                f"Different frame rates detected ({fps}). "
                f"Setting the frame rate to the mean value: {fps[0]} fps"
            )

        return widths[0], heights[0], fps[0]

    # Methods to create folders where to store data
    # TODO: Some of these methods should go to the classes corresponding to
    # the process.

    def create_accumulation_folder(self, iteration_number=None, delete=False):
        """Folder in which the model generated while accumulating is stored
        (after pretraining)
        """
        if iteration_number is None:
            iteration_number = self.accumulation_trial
        self.accumulation_folder = (
            self.session_folder / f"accumulation_{iteration_number}"
        )
        # FIXME
        create_dir(self.accumulation_folder, remove_existing=delete)

    # Some methods related to the accumulation process
    # TODO: Move to accumulation_manager.py
    def init_accumulation_statistics_attributes(self):
        self.accumulation_statistics = {
            "n_accumulated_global_fragments": [],
            "n_non_certain_global_fragments": [],
            "n_randomly_assigned_global_fragments": [],
            "n_nonconsistent_global_fragments": [],
            "n_nonunique_global_fragments": [],
            "n_acceptable_global_fragments": [],
            "ratio_of_accumulated_images": [],
        }

    @staticmethod
    def get_processing_episodes(
        video_paths, frames_per_episode, tracking_intervals=None
    ) -> tuple[(int, list[int], list[list[int]], list[Episode])]:
        """Process the episodes by getting the number of frames in each video
        path and the tracking interval.

        Episodes are used to compute processes in parallel for different
        parts of the video. They are a tuple with
            (local start frame,
            local end frame,
            video path index,
            global start frame,
            global end frame)
        where "local" means in the specific video path and "global" means in
        the whole (multi path) video

        Episodes are guaranteed to belong to a single video path and to have
        all of their frames (end not included) inside a the tracking interval
        """

        def in_which_interval(frame_number, intervals) -> int | None:
            for i, (start, end) in enumerate(intervals):
                if start <= frame_number < end:
                    return i
            return None

        # total number of frames for every video path
        video_paths_n_frames = [
            int(cv2.VideoCapture(str(path)).get(7)) for path in video_paths
        ]
        number_of_frames = sum(video_paths_n_frames)

        # set full tracking interval if not defined
        if tracking_intervals is None:
            tracking_intervals = [[0, number_of_frames]]
        elif isinstance(tracking_intervals[0], int):
            tracking_intervals = [tracking_intervals]

        # find the global frames where the video path changes
        video_paths_changes = [0] + list(np.cumsum(video_paths_n_frames))

        # build an interval list like ("frame" refers to "global frame")
        #   [[first frame of video path 0, last frame of video path 0],
        #    [first frame of video path 1, last frame of video path 1],
        #    [...]]
        video_paths_intervals = list(pairwise(video_paths_changes))

        # find the frames where a tracking interval starts or ends
        tracking_intervals_changes = list(np.asarray(tracking_intervals).flatten())

        # Take into account tracking interval changes
        # and video path changes to compute episodes
        limits = video_paths_changes + tracking_intervals_changes

        # clean repeated limits and sort them
        limits = sorted(set(limits))

        # Create "long episodes" as the intervals between any video path
        # change or tracking interval change (keeping only the ones that
        # are inside a tracking interval)
        long_episodes = []
        for start, end in pairwise(limits):
            if (
                in_which_interval(start, tracking_intervals) is not None
            ) and 0 <= start < number_of_frames:
                long_episodes.append((start, end))

        # build definitive episodes by dividing long episodes to fit in
        # the FRAMES_PER_EPISODE restriction
        index = 0
        episodes = []
        for start, end in long_episodes:
            video_path_index = in_which_interval(start, video_paths_intervals)
            assert video_path_index is not None
            global_local_offset = video_paths_intervals[video_path_index][0]

            n_subepisodes = int((end - start) / (frames_per_episode + 1))
            new_episode_limits = np.linspace(start, end, n_subepisodes + 2, dtype=int)
            for new_start, new_end in pairwise(new_episode_limits):
                episodes.append(
                    Episode(
                        index=index,
                        local_start=new_start - global_local_offset,
                        local_end=new_end - global_local_offset,
                        video_path_index=video_path_index,
                        global_start=new_start,
                        global_end=new_end,
                    )
                )
                index += 1
        return number_of_frames, video_paths_n_frames, tracking_intervals, episodes

    @staticmethod
    def in_which_interval(frame_number, intervals):
        for i, (start, end) in enumerate(intervals):
            if start <= frame_number < end:
                return i
        return None

    def delete_data(self):
        """Deletes some folders with data, to make the outcome lighter.

        Which folders are deleted depends on the constant DATA_POLICY
        """

        logging.info(f'Data policy: "{self.data_policy}"')

        if self.data_policy == "trajectories":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            for path in self.session_folder.glob("accumulation_*"):
                remove_dir(path)
            remove_dir(self.session_folder / "pretraining")
            remove_dir(self.preprocessing_folder)
        elif self.data_policy == "validation":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            for path in self.session_folder.glob("accumulation_*"):
                remove_dir(path)
            remove_dir(self.session_folder / "pretraining")
        elif self.data_policy == "knowledge_transfer":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
        elif self.data_policy == "idmatcher.ai":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)

    def compress_data(self):
        """Compress the identification images h5py files"""
        if not self.id_images_folder.exists():
            return

        tmp_path = self.session_folder / "tmp.h5py"

        for path in track(
            self.id_images_file_paths, "Compressing identification images"
        ):
            if not path.is_file():
                continue
            with (
                h5py.File(path, "r") as original_file,
                h5py.File(tmp_path, "w") as compressed_file,
            ):
                for key, data in original_file.items():
                    compressed_file.create_dataset(
                        key, data=data, compression="gzip" if "image" in key else None
                    )
            path.unlink()  # Windows needs this call before rename()
            tmp_path.rename(path)



from qtpy.QtCore import QRectF, Qt
from qtpy.QtGui import QColor, QColorConstants, QPainter, QPaintEvent
from qtpy.QtWidgets import QWidget


class BlobInfoWidget(QWidget):
    bar_width = 0.65

    def __init__(self):
        super().__init__()
        self.areas = []
        self.frame = 0
        self.n_animals = 0
        self.tracking_intervals: list[list[int]] | None = None
        self.setMinimumSize(100, 100)

    def in_tracking_intervals(self, frame) -> bool:
        if self.tracking_intervals is None:
            return True
        return any(start <= frame < end for start, end in self.tracking_intervals)

    def setAreas(self, frame: int, areas: list[int]):
        self.frame = frame
        self.areas = areas
        self.update()

    def setNAnimals(self, n_animals):
        self.n_animals = n_animals
        self.update()

    def setTrackingIntervals(self, tracking_intervals):
        self.tracking_intervals = tracking_intervals
        self.update()

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        base_color = painter.pen().color()
        left_margin = 8
        w = self.width() - left_margin
        h = self.height()

        if w > 700:
            middle = int(0.5286 * w)
            left = middle - 300 + left_margin
            right = middle + 300
        else:
            left = 70 + left_margin
            right = w - 30
        axis_w = right - left

        if h > 500:
            middle = h // 2
            top = middle - 200
            bottom = middle + 220
        else:
            top = 50
            bottom = h - 30
        axis_h = bottom - top

        painter.drawText(
            0,
            bottom + 5,
            w,
            bottom + 50,
            Qt.AlignmentFlag.AlignHCenter,
            "Detected blobs",
        )
        painter.save()
        painter.translate(0, h)
        painter.rotate(-90)
        painter.drawText(
            0,
            0,
            h,
            left - 50,
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignBottom,
            "Area (px)",
        )
        painter.restore()

        number_of_blobs = len(self.areas)

        scale = 1
        rects = []
        if not self.isEnabled():
            title = ""
            min_area_line = None
        elif not self.in_tracking_intervals(self.frame):
            title = "Frame outside tracking intervals"
            min_area_line = None
        elif number_of_blobs == 0:
            title = "No blobs detected"
            min_area_line = None
        else:
            if self.n_animals == 0:
                title_prefix = "# animals = 0! "
                painter.setBrush(QColor(0xBA2320))
                painter.setPen(QColor(0x5A1010))
            elif number_of_blobs > self.n_animals:
                title_prefix = "More blobs than animals! "
                painter.setBrush(QColor(0xBA2320))
                painter.setPen(QColor(0x5A1010))
            else:
                title_prefix = ""
                painter.setBrush(QColor(0x44A0D9))
                painter.setPen(QColor(0x286384))

            bar_sep = axis_w / number_of_blobs
            bar_width = 0.7 * axis_w / number_of_blobs
            scale = axis_h / (1.1 * max(self.areas))
            rects = [
                QRectF(
                    left + (i + 0.5) * bar_sep - 0.5 * bar_width,
                    bottom - area * scale,
                    bar_width,
                    area * scale,
                )
                for i, area in enumerate(self.areas)
            ]
            painter.drawRects(rects)  # type: ignore

            min_area_line = min(self.areas)
            if number_of_blobs == 1:
                title = f"1 blob detected of area {self.areas[0]:.0f} px"
            else:
                title = (
                    title_prefix
                    + f"{number_of_blobs} blobs detected. "
                    f"Minimum area: {min_area_line:.0f} px"
                )

        # Draw min area dashed line
        if min_area_line is not None:
            pen = painter.pen()
            pen.setColor(QColor(0x808080))
            pen.setStyle(Qt.PenStyle.DotLine)
            pen.setWidth(2)
            painter.setPen(pen)
            painter.drawLine(
                left,
                bottom - int(min_area_line * scale),
                right,
                bottom - int(min_area_line * scale),
            )

        # Draw title
        painter.setPen(
            QColorConstants.Red if title == "Number of animals missing!" else base_color
        )
        painter.drawText(
            0,
            0,
            w,
            top,
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
            title,
        )

        # Draw ticks
        painter.setPen(base_color)
        tick_lenght = 10
        if rects:
            lim = 1.1 * max(self.areas)
            for frequency in (
                1,
                2,
                5,
                10,
                25,
                50,
                100,
                250,
                500,
                100,
                2500,
                5000,
                10000,
                25000,
                50000,
                100000,
                250000,
            ):
                n_ticks = int(lim / frequency)
                if n_ticks < 5:
                    break
            for tick_index in range(n_ticks + 1):
                tick_value = frequency * tick_index
                tick_high = int(bottom - tick_value * scale)
                painter.drawLine(left - tick_lenght, tick_high, left, tick_high)
                painter.drawText(
                    0,
                    tick_high - 20,
                    left - tick_lenght - 3,
                    40,
                    Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter,
                    f"{tick_value}",
                )

        # Draw axes
        painter.drawLine(left, bottom, right, bottom)
        painter.drawLine(left, bottom, left, top)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)


import numpy as np

from idtrackerai import ListOfFragments
from idtrackerai.utils import conf, track


def compute_model_velocity(
    list_of_fragments: ListOfFragments, percentile=None
) -> float:
    """computes the 2 * (percentile) of the distribution of velocities of identified fish.
    params
    -----
    blobs_in_video: list of blob objects
        collection of blobs detected in the video.
    number_of_animals int
    percentile int
    -----
    return
    -----
    float
    2 * np.max(distance_travelled_in_individual_fragments) if percentile is None
    2 * percentile(velocity distribution of identified animals) otherwise
    """
    if percentile is None:
        percentile = conf.VEL_PERCENTILE
    distance_travelled_in_individual_fragments: list[float] = []

    for fragment in track(
        list_of_fragments.individual_fragments, "Computing velocity model"
    ):
        distance_travelled_in_individual_fragments += (
            fragment.frame_by_frame_velocity.tolist()
        )
    return (
        2 * np.max(distance_travelled_in_individual_fragments)
        if percentile is None
        else 2.0
        * float(np.percentile(distance_travelled_in_individual_fragments, percentile))
    )



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
from functools import cached_property
from itertools import chain
from math import atan2, sqrt
from pathlib import Path
from typing import Iterable, Sequence

import cv2
import h5py
import numpy as np


class Blob:
    """Represents a segmented blob (collection of pixels) from a given frame.

    A blob can represent a single animal or multiple animals during an
    occlusion or crossing.

    Parameters
    ----------

    contour : numpy array
        Array with the points that define the contour of the blob. The
        array is of the form [[[x1,y1]],[[x2,y2]],...,[[xn,yn]]].
    estimated_body_length : float, optional
        Body length of the animal estimated from the diagonal of the
        original bounding box, by default None.
    pixels : list, optional
        List of pixels that define the blob, by default None.
    number_of_animals : int, optional
        Number of animals in the video as defined by the user,
        by default None.
    frame_number : int, optional
        Frame number in the video from which the blob was segmented,
        by default None.
    frame_number_in_video_path : int, optional
        Frame number in the video path from which the blob was segmented,
        by default None.
    in_frame_index : int, optional
        Index of the blob in the frame where it was segmented,
        by default None. This index comes from OpenCV and is defined by the
        hierarchy of the countours found in the frame.
    pixels_are_from_eroded_blob : bool, optional
        Flag to indicate if the pixels of the blobs come from from an
        eroded blob, by default False.
    resolution_reduction : float, optional
        Resolution reductio factor as defined by the user, by default 1.0.
    """

    episode: int
    id_image_index: int
    """Index of the identification image position in the hdf5 file"""

    seems_like_individual: bool = False
    """Unicity condition or not huge area"""

    is_an_individual: bool
    """Flag indicating the blob represents a single animal.
    Defined in crossing detection."""

    used_for_training_crossings: bool = False
    """Flag indicating if the blob has been used to train the
    crossing CNN"""

    was_a_crossing: bool = False
    """Flag indicating whether the blob was created after splitting a
    crossing blob during the crossings interpolation process"""

    identity: int | None = None
    """Identity of the blob assigned during the identification process"""

    next: list["Blob"]
    previous: list["Blob"]

    fragment_identifier: int = -1
    """Indicates the index of the Fragment that contains the blob,
    -1 means no associated Fragment"""

    pixels_are_from_eroded_blob: bool = False

    blob_index: int = None  # type: ignore
    """Blob index at the segmentation step (comes from the find contours
    function of OpenCV)"""

    user_generated_identities: list[int | None] = None  # type: ignore
    """List of identities of the blob some of which might have been give
    by a user during the validation process"""

    user_generated_identity: int | None = None
    """This property is give during the correction of impossible velocity
    jumps. It has nothing to do with the manual validation."""

    user_generated_centroids: list[tuple[float, float] | None] = None  # type: ignore
    """List of centroids generated by the user during the validation
    processes"""

    identity_corrected_solving_jumps: int | None = None
    """Identity of the blob after correcting impossible velocity jumps"""

    identities_corrected_closing_gaps: list | None = None
    """Identity of the blob after crossings interpolation"""

    interpolated_centroids: list | None = None

    added_by_user: bool = False

    forced_crossing: bool = False
    """Indicates if the crossing attribute has been forced by set_individual_with_identity_0_as_crossings()"""

    def __init__(
        self,
        contour: np.ndarray,
        frame_number: int = -1,
        bbox_img_id: str = "",
        pixels_are_from_eroded_blob: bool = False,
    ):
        self.set_contour(contour)
        self.frame_number = frame_number
        self.bbox_img_id = bbox_img_id
        if pixels_are_from_eroded_blob:
            self.pixels_are_from_eroded_blob = pixels_are_from_eroded_blob
            # TODO fix this, some eroded blobs are not classified as
            # individuals/crossings and idtrackerai crashes
            self.is_an_individual = True
            self.forced_crossing = True

        self.next = []
        self.previous = []

    @property
    def n_next(self):
        return len(self.next)

    @property
    def n_previous(self):
        return len(self.previous)

    @cached_property
    def convexHull(self) -> np.ndarray:
        return cv2.convexHull(self.contour)

    @cached_property
    def area(self) -> int:
        return cv2.contourArea(self.contour)

    @cached_property
    def bbox_in_frame_coordinates(self) -> tuple[tuple[int, int], tuple[int, int]]:
        return self.contour.min(0), self.contour.max(0)

    @property
    def estimated_body_length(self):
        width, height = np.ptp(self.contour, axis=0)
        return int(np.ceil(sqrt(width**2 + height**2)))

    @cached_property
    def centroid(self) -> tuple[float, float]:
        M = cv2.moments(self.contour)
        try:
            return M["m10"] / M["m00"], M["m01"] / M["m00"]
        except ZeroDivisionError:
            return tuple(np.mean(self.contour, axis=0))

    @cached_property
    def orientation(self) -> float:
        M = cv2.moments(self.contour)
        try:
            x = M["m10"] / M["m00"]
            y = M["m01"] / M["m00"]
            a = M["m20"] / M["m00"] - x * x
            b = 2 * (M["m11"] / M["m00"] - x * y)
            c = M["m02"] / M["m00"] - y * y
            return 0.5 * atan2(b, (a - c))
        except ZeroDivisionError:
            return 0

    def set_contour(self, contour: np.ndarray):
        if contour.ndim == 3 and contour.shape[1] == 1:
            # OpenCV returns contours as (n_points, 1, 2)
            contour = contour[:, 0, :]
        self.contour = contour.astype(np.int32, copy=False)

    @property
    def is_a_crossing(self) -> bool:
        """Flag indicating whether the blob represents two or more animals
        together.

        This attribute is the negative of `is_an_individual` and is set at
        the same time as is an individual
        """
        return not self.is_an_individual

    def has_multiple_previous(self) -> bool:
        """Flag indicating if the blob has multiple blobs in its past or future
        overlapping history

        This method is used to check whether the blob is a crossing.

        Returns
        -------
        bool
            If True the blob splits into two or multiple overlapping blobs in
            its "past" or "future" history, depending on the parameter
            "direction".
        """

        current = self.previous[0]
        while len(current.previous) == 1:
            current = current.previous[0]
            if len(current.previous) > 1:
                return True

        return False

    def has_multiple_next(self) -> bool:
        """Flag indicating if the blob has multiple blobs in its past or future
        overlapping history

        This method is used to check whether the blob is a crossing.

        Returns
        -------
        bool
            If True the blob splits into two or multiple overlapping blobs in
            its "past" or "future" history, depending on the parameter
            "direction".
        """

        current = self.next[0]
        while len(current.next) == 1:
            current = current.next[0]
            if len(current.next) > 1:
                return True

        return False

    def check_for_crossing_next(self) -> bool:
        """Flag indicating if the blob has a crossing in its future overlapping history

        Returns
        -------
        bool
            If True the blob has a crossing in its "future" history
        """
        current = self.next[0]
        while current.n_next == 1:
            current = current.next[0]
            if current.n_previous > 1 and not current.seems_like_individual:
                return True
        return False

    def check_for_crossing_previous(self) -> bool:
        """Flag indicating if the blob has a crossing in its past overlapping history

        Returns
        -------
        bool
            If True the blob has a crossing in its "past" history
        """
        current = self.previous[0]
        while current.n_previous == 1:
            current = current.previous[0]
            if current.n_next > 1 and not current.seems_like_individual:
                return True
        return False

    def is_a_sure_individual(self) -> bool:
        """Flag indicating that the blob is a sure individual according to
        some heuristics and it can be used to train the crossing detector CNN.
        """
        return (
            self.seems_like_individual
            and len(self.previous) == 1
            and len(self.next) == 1
            and len(self.next[0].previous) == 1
            and len(self.previous[0].next) == 1
            and self.check_for_crossing_previous()
            and self.check_for_crossing_next()
        )

    def is_a_sure_crossing(self) -> bool:
        """Flag indicating that the blob is a sure crossing according to
        some heuristics and it can be used to train the crossing detector CNN.

        Returns
        -------
        bool
        """
        if self.seems_like_individual:
            return False
        if len(self.previous) > 1 or len(self.next) > 1:
            return True
        if len(self.previous) == 1 and len(self.next) == 1:
            return self.has_multiple_previous() and self.has_multiple_next()
        return False

    def overlaps_with(self, other: "Blob") -> bool:
        """Computes whether the pixels in `self` intersect with the pixels in
        `other`

        Parameters
        ----------
        other : <Blob object>
            An instance of the class Blob

        Returns
        -------
        bool
            True if the lists of pixels of both blobs have non-empty
            intersection
        """

        # Check bounding boxes
        (S_xmin, S_ymin), (S_xmax, S_ymax) = self.bbox_in_frame_coordinates
        (O_xmin, O_ymin), (O_xmax, O_ymax) = other.bbox_in_frame_coordinates

        if not S_xmax >= O_xmin and O_xmax >= S_xmin:  # x overlap
            return False
        if not S_ymax >= O_ymin and O_ymax >= S_ymin:  # y overlap
            return False

        # Check convex hull
        if not cv2.intersectConvexConvex(self.convexHull, other.convexHull)[0]:
            return False

        # Check for every point in `other`'s contour
        points = other.contour.astype(float)
        for point in chain(points[0::3], points[1::3], points[2::3]):
            if self.contour_contains_point(point):
                return True

        # Check for every point in `self`'s contour
        points = self.contour.astype(float)
        return any(
            other.contour_contains_point(point)
            for point in chain(points[0::3], points[1::3], points[2::3])
        )

    def contour_contains_point(self, point: tuple[float, float]) -> bool:
        return cv2.pointPolygonTest(self.contour, point, False) >= 0

    def bbox_contains_point(self, point: tuple[float, float]) -> bool:
        return (
            point[0] >= self.bbox_in_frame_coordinates[0][0]
            and point[0] <= self.bbox_in_frame_coordinates[1][0]
            and point[1] >= self.bbox_in_frame_coordinates[0][1]
            and point[1] <= self.bbox_in_frame_coordinates[1][1]
        )

    def contains_point(self, point: tuple[float, float]) -> bool:
        if not self.bbox_contains_point(point):
            return False
        return self.contour_contains_point(point)

    def now_points_to(self, other: "Blob"):
        """Given two consecutive blob objects updates their respective
        overlapping histories

        Parameters
        ----------
        other : <Blob object>
            An instance of the class Blob
        """
        self.next.append(other)
        other.previous.append(self)

    def square_distance_to(self, other: "Blob|tuple|list|np.ndarray"):
        """Returns the squared distance from the centroid of self to the
        centroid of `other`

        Parameters
        ----------
        other : <Blob object> or tuple
            An instance of the class Blob or a tuple (x,y)

        Returns
        -------
        float
            Squared distance between centroids
        """
        if isinstance(other, Blob):
            return ((np.asarray(self.centroid) - np.asarray(other.centroid)) ** 2).sum()

        if isinstance(other, (tuple, list, np.ndarray)):
            return ((np.asarray(self.centroid) - np.asarray(other)) ** 2).sum()

        raise ValueError

    def distance_to(self, other: "Blob|tuple|list|np.ndarray") -> float:
        return sqrt(self.square_distance_to(other))

    def distance_from_countour_to(self, point):
        """Returns the distance between `point` and the closest
        point in the contour of the blob.

        Parameters
        ----------
        point : tuple
            (x,y)

        Returns
        -------
        float
            Smallest distance between `point` and the contour of the blob.
        """
        return abs(cv2.pointPolygonTest(self.contour, point, True))

    @property
    def assigned_identities(self):
        """Identities assigned to the blob during the tracking process"""
        if self.identities_corrected_closing_gaps is not None:
            return self.identities_corrected_closing_gaps
        if self.identity_corrected_solving_jumps is not None:
            return [self.identity_corrected_solving_jumps]
        return [self.identity]

    @property
    def assigned_centroids(self):
        """Centroids assigned to the blob during the tracking process.

        It considers the default centroid of the blob at segmentation time
        or new centroids added to the blob during the interpolation of the
        crossings.

        Returns
        -------
        list
            List of pairs (x, y) indicating the position of each individual
            in the blob.
        """
        if self.interpolated_centroids:
            return self.interpolated_centroids
        return [self.centroid]

    @property
    def final_centroids(self):
        """List of the animal/s centroid/s in the blob, considering the
        potential centroids that might have been added by the user during
        the validation.

        By default the centroid will be the center of mass of the blob of
        pixels defined by the blob. It can be different if the user modified
        the default centroid during validation or generated more centroids.

        Returns
        -------
        list
            List of tuples (x, y) indicating the centroids of the blob.
        """
        return (c for c in self.all_final_centroids if c != (-1, -1))

    @property
    def all_final_centroids(self):
        if self.user_generated_centroids:
            # Note that sometimes len(user_generated_centroids) >
            # len(assigned_centroids)
            final_centroids = []
            for i, centroid in enumerate(self.user_generated_centroids):
                if centroid is not None or i >= len(self.assigned_centroids):
                    final_centroids.append(centroid)
                else:
                    final_centroids.append(self.assigned_centroids[i])
            return final_centroids
        return self.assigned_centroids

    @property
    def final_identities(self):
        return (id for id in self.all_final_identities if id != -1)

    @property
    def all_final_identities(self):
        """Identities of the blob after the tracking process and after
        potential modifications by the users during the validation procedure.
        """
        if self.user_generated_identities:
            # Note that sometimes len(user_generated_identities)
            # > len(assigned_identities)
            final_identities = []
            # TODO None means the same as assigned, 0 means no id, -1 means no centroid
            for i, user_generated_identity in enumerate(self.user_generated_identities):
                if user_generated_identity is not None or i >= len(
                    self.assigned_identities
                ):
                    final_identities.append(user_generated_identity)
                else:
                    final_identities.append(self.assigned_identities[i])
            return final_identities
        return self.assigned_identities

    @property
    def final_ids_and_centroids(
        self,
    ) -> Iterable[tuple[int | None, tuple[float, float]]]:
        return zip(self.final_identities, self.final_centroids)

    @property
    def all_final_ids_and_centroids(self):
        return zip(self.all_final_identities, self.all_final_centroids)

    def get_image_for_identification(
        self, img_size: int, bbox_imgs_path: Path
    ) -> np.ndarray:
        """Gets the image used to train and evaluate the crossing detector CNN
        and the identification CNN.

        Parameters
        ----------
        id_image_size : tuple
            Dimensions of the identification image (height, widht, channels).
            Channels is always 1 as images in color are still not considered.
        height : int
            Video height considering resolution reduction factor.
        width : int
            Video width considering resolution reduction factor.

        Returns
        -------
        ndarray
            Square image with black background used to train the crossings
            detector CNN and the identifiactio CNN.

        It generates the image that will be used to train and evaluate the
        crossings detector CNN and the identification CNN.

        Parameters
        ----------
        height : int
            Frame height
        width : int
            Frame width
        bbox_image : ndarray
            Images cropped from the frame by considering the bounding box
            associated to a blob
        pixels : list
            List of pixels associated to a blob
        bbox_in_frame_coordinates : list
            [(x, y), (x + bbox_width, y + bbox_height)]
        image_size : int
            Size of the width and height of the square identification image

        Returns
        -------
        ndarray
            Square image with black background used to train the crossings
            detector CNN and the identification CNN.
        """

        with h5py.File(bbox_imgs_path, "r") as f:
            bbox_img = f[self.bbox_img_id][:]  # type: ignore #

        mask = self.get_bbox_mask()

        mask = cv2.dilate(mask, np.ones((3, 3), np.uint8), iterations=1)

        masked_bbox_image = bbox_img * mask
        bbox_img_height, bbox_img_width = masked_bbox_image.shape
        img_size2 = img_size % 2 + img_size // 2

        center_x = int(
            self.centroid[0]
            - self.bbox_in_frame_coordinates[0][0]
            + 1  # bbox_image_pad
        )

        center_y = int(
            self.centroid[1]
            - self.bbox_in_frame_coordinates[0][1]
            + 1  # bbox_image_pad
        )

        d1 = center_x**2 + center_y**2
        d2 = center_x**2 + (bbox_img_height - center_y) ** 2
        d3 = (bbox_img_width - center_x) ** 2 + center_y**2
        d4 = (bbox_img_width - center_x) ** 2 + (bbox_img_height - center_y) ** 2
        diag = int(sqrt(max((d1, d2, d3, d4))))
        diag = max(diag, img_size2)
        id_img = np.zeros((2 * diag, 2 * diag), np.uint8)
        id_img[
            diag - center_y : diag + bbox_img_height - center_y,
            diag - center_x : diag + bbox_img_width - center_x,
        ] = masked_bbox_image

        M = cv2.getRotationMatrix2D(
            (diag, diag), self.orientation * 180 / np.pi - 45, 1
        )

        # old method
        id_img = cv2.warpAffine(
            src=id_img,
            M=M,
            dsize=(diag + img_size2, diag + img_size2),
            borderMode=cv2.BORDER_CONSTANT,
            flags=cv2.INTER_CUBIC,
        )
        return id_img[-img_size:, -img_size:]

        # TODO proposed future method
        # id_img = cv2.warpAffine(
        #     src=id_img,
        #     M=M,
        #     dsize=(diag + img_size, diag + img_size),
        #     borderMode=cv2.BORDER_CONSTANT,
        #     flags=cv2.INTER_CUBIC,
        # )

        # # we build the offset like this to have the minimal ones on the
        # # beginning of the array and be preferably selected by max()
        # origins = [0]
        # for offset in range(img_size2 // 2):
        #     origins += (diag - img_size2 + offset, diag - img_size2 - offset)

        # origin = max(
        #     origins,
        #     key=lambda origin: np.count_nonzero(
        #         id_img[origin : origin + img_size, origin : origin + img_size]
        #     ),
        # )

        # return id_img[origin : origin + img_size, origin : origin + img_size]

    def get_bbox_mask(self) -> np.ndarray:
        base = np.zeros(
            (
                self.bbox_in_frame_coordinates[1][1]
                - self.bbox_in_frame_coordinates[0][1]
                + 2,  # 2 bbox_image_pads
                self.bbox_in_frame_coordinates[1][0]
                - self.bbox_in_frame_coordinates[0][0]
                + 2,  # 2 bbox_image_pads
            ),
            np.uint8,
        )
        return cv2.fillPoly(
            img=base,
            pts=(self.contour,),
            color=1,
            offset=(
                1 - self.bbox_in_frame_coordinates[0][0],  # bbox_image_pad
                1 - self.bbox_in_frame_coordinates[0][1],  # bbox_image_pad
            ),
        )

    def update_centroid(
        self,
        old_centroid: tuple[float, float],
        new_centroid: tuple[float, float],
        identity: int,
    ):
        """[Validation] Updates the centroid of the blob.

        Parameters
        ----------
        video : idtrackerai.video.Video
            Instance of Video object
        old_centroid : tuple
            Centroid to be updated
        new_centroid : tuple
            Coordinates of the new centroid
        identity : int
            Identity of the centroid to be updated
        """

        self.init_validator_variables()
        index, centroid, dist = self.index_and_centroid_closer_to(
            old_centroid, identity
        )

        self.user_generated_centroids[index] = new_centroid
        self.user_generated_identities[index] = identity

    def init_validator_variables(self):
        if self.user_generated_centroids is None:
            self.user_generated_centroids: list[tuple[float, float] | None] = [
                None
            ] * len(list(self.final_centroids))
        if self.user_generated_identities is None:
            self.user_generated_identities = [None] * len(list(self.final_identities))

    def index_and_centroid_closer_to(
        self, centroid: tuple, identity: int | None
    ) -> tuple[int, tuple, float]:
        candidates: list[tuple[int, tuple, float]] = []
        for indx, (_id, _centroid) in enumerate(self.all_final_ids_and_centroids):
            if identity not in (None, _id):
                continue
            dist = (_centroid[0] - centroid[0]) ** 2 + (_centroid[1] - centroid[1]) ** 2
            candidates.append((indx, _centroid, dist))
        if not candidates:
            raise ValueError("Centroid not found")

        return min(candidates, key=lambda x: x[0])

    def remove_centroid(self, identity: int, centroid: tuple):
        """[Validation] Deletes a centroid of the blob.

        Parameters
        ----------
        identity : int
            Identity of the centroid to be deleted
        centroid : tuple
            Centroid to be deleted from the blob
        """

        self.init_validator_variables()
        index, centroid, dist = self.index_and_centroid_closer_to(centroid, identity)

        self.user_generated_centroids[index] = (-1, -1)
        self.user_generated_identities[index] = -1

    def add_centroid(self, centroid: tuple[float, float], identity: int | None):
        """[Validation] Adds a centroid with a given identity to the blob.

        This method is used in the validation GUI. It is useful to add
        centroids for crossing blobs that are missing some centroids, or to
        individual blobs that should have been classified as crossings and
        are also missing some centroids.

        Parameters
        ----------
        centroid : tuple
            Centroid to be added to the blob
        identity : int
            Identity of the centroid to be added
        """

        self.init_validator_variables()

        self.user_generated_centroids.append(centroid)
        self.user_generated_identities.append(identity)

    def update_identity(
        self,
        old_identity: int | None,
        new_identity: int | None,
        close_to_centroid: tuple[float, float],
    ):
        """[Validation] Updates the identity of the blob.

        This method is used during the validation GUI.
        It populates the private attributes `user_generated_identities`
        and `user_generated_centroids`.

        Parameters
        ----------
        new_identity : int
            new value for the identity of the blob
        old_identity : int
            old value of the identity of the blob. It must be specified when the
            blob has multiple identities already assigned.
        centroid : tuple
            centroid which identity must be updated.
        """
        self.init_validator_variables()

        index, centroid, dist = self.index_and_centroid_closer_to(
            close_to_centroid, old_identity
        )

        self.user_generated_identities[index] = new_identity
        self.user_generated_centroids[index] = centroid
        return centroid

    def propagate_identity(
        self,
        old_identity: int | None,
        new_identity: int | None,
        centroid: tuple[float, float],
    ):
        """[Validation] Propagates the new identity to next and previous blobs.

        This method called in the validation GUI when the used updates the
        identity of a given blob.

        Returns the frame range where identity changes occurred

        Parameters
        ----------
        old_identity : int
            Previous identity of the blob
        new_identity : int
            New identity of the blob
        centroid : tuple
            [description]
        """

        blobs_stack: list[tuple[Blob, tuple[float, float]]]
        first_frame_modified = self.frame_number
        last_frame_modified = self.frame_number

        blobs_stack = [(next_blob, centroid) for next_blob in self.next]
        while blobs_stack:
            current, previous_centroid = blobs_stack.pop()
            if current.fragment_identifier != self.fragment_identifier:
                continue
            try:
                new_centroid = current.update_identity(
                    old_identity, new_identity, previous_centroid
                )
            except ValueError:  # centroid not found on "current"
                continue
            blobs_stack += [(next_blob, new_centroid) for next_blob in current.next]
            last_frame_modified = current.frame_number

        blobs_stack += [(prev_blob, centroid) for prev_blob in self.previous]
        while blobs_stack:
            current, previous_centroid = blobs_stack.pop()
            if current.fragment_identifier != self.fragment_identifier:
                continue

            try:
                new_centroid = current.update_identity(
                    old_identity, new_identity, previous_centroid
                )
            except ValueError:  # centroid not found on "current"
                continue
            blobs_stack += [(prev_blob, new_centroid) for prev_blob in current.previous]
            first_frame_modified = current.frame_number

        return first_frame_modified, last_frame_modified

    @property
    def properties(self) -> Sequence[str]:
        return (
            (
                (("Individual" if self.is_an_individual else "Crossing") + " Blob")
                + (" (forced)" if self.forced_crossing else "")
            ),
            f"{self.contour.shape[0]} vertices in contour of {self.area:.0f} px area",
            f"In fragment {self.fragment_identifier}",
            f"Linked to {self.n_previous} previous blobs",
            f"Linked to {self.n_next} next blobs",
            ("Used" if self.used_for_training_crossings else "Not used")
            + " for training crossings",
            f"Seems like individual: {self.seems_like_individual}",
            f"Pixels are from eroded blob: {self.pixels_are_from_eroded_blob}",
            f"Predicted identity: {self.identity}",
            f"Corrected solving jumps {self.identity_corrected_solving_jumps}",
            f"Corrected solving gaps: {self.identities_corrected_closing_gaps}",
            f"assigned identities: {self.assigned_identities}",
            f"assigned centroids: {repr_of_list_of_points(self.assigned_centroids)}",
            f"user identities: {self.user_generated_identities}",
            f"user centroids: {repr_of_list_of_points(self.user_generated_centroids)}",
            f"final identities: {list(self.final_identities)}",
            f"final centroids: {repr_of_list_of_points(self.final_centroids)}",
        )


def repr_of_list_of_points(list_of_points) -> str:
    if list_of_points is None:
        return "None"
    list_of_str = [
        f"({point[0]:.1f}, {point[1]:.1f})" if point is not None else "None"
        for point in list_of_points
    ]
    return f"[{', '.join(list_of_str)}]"



import logging

import numpy as np
import torch
from torch.backends import cudnn

from idtrackerai import ListOfBlobs, ListOfFragments, ListOfGlobalFragments, Video
from idtrackerai.network import (
    DEVICE,
    LearnerClassification,
    NetworkParams,
    fc_weights_reinit,
    weights_xavier_init,
)
from idtrackerai.utils import CustomError, conf, create_dir

from .accumulation_manager import AccumulationManager
from .accumulator import perform_one_accumulation_step
from .assigner import assign_remaining_fragments
from .identity_transfer import identify_first_global_fragment_for_accumulation
from .pre_trainer import pretrain_global_fragment


class TrackerAPI:
    identification_model: torch.nn.Module
    accumulation_network_params: NetworkParams

    def __init__(
        self,
        video: Video,
        list_of_blobs: ListOfBlobs,
        list_of_fragments: ListOfFragments,
        list_of_global_fragments: ListOfGlobalFragments,
    ):
        self.video = video
        self.list_of_blobs = list_of_blobs
        self.list_of_fragments = list_of_fragments
        self.list_of_global_fragments = list_of_global_fragments

    def track_single_animal(self):
        logging.debug("Assigning identity 1 to all blobs")
        for blob in self.list_of_blobs.all_blobs:
            blob.identity = 1

    def track_single_global_fragment_video(self):
        logging.info("TRACKING SINGLE GLOBAL FRAGMENT")
        assert len(self.list_of_global_fragments.global_fragments) == 1
        global_fragment = self.list_of_global_fragments.global_fragments[0]

        for identity, fragment in enumerate(global_fragment.individual_fragments):
            fragment.identity = identity + 1

        self.list_of_fragments.update_blobs(self.list_of_blobs.all_blobs)

    def track_with_identities(self) -> ListOfFragments:
        """In protocol 3, list_of_fragments is loaded from accumulation
        folders so the reference from outside tracker_API is lost.
        That's why list_of_fragments has to be returned"""
        self.video.tracking_timer.start()
        self.track_with_protocols_cascade()
        self.video.tracking_timer.finish()
        return self.list_of_fragments

    def track_with_protocols_cascade(self):
        logging.info("Starting protocol cascade")
        self.video.create_accumulation_folder(iteration_number=0, delete=True)
        self.accumulation_network_params = NetworkParams(
            n_classes=self.video.n_animals,
            architecture=conf.IDCNN_NETWORK_NAME,
            save_folder=self.video.accumulation_folder,
            knowledge_transfer_folder=self.video.knowledge_transfer_folder,
            model_name="identification_network",
            image_size=self.video.id_image_size,
            scopes_layers_to_optimize=conf.LAYERS_TO_OPTIMISE_PRETRAINING,
            optimizer="SGD",
            schedule=[30, 60],
            optim_args={"lr": conf.LEARNING_RATE_IDCNN_ACCUMULATION, "momentum": 0.9},
            epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
        )
        self.accumulation_network_params.save()
        self.protocol1()

    def protocol1(self):
        self.video.protocol1_timer.start()

        # reset list of fragments and global fragments to fragmentation
        self.list_of_fragments.reset(roll_back_to="fragmentation")

        if self.video.knowledge_transfer_folder:
            try:
                self.identification_model = LearnerClassification.load_model(
                    self.accumulation_network_params, knowledge_transfer=True
                )
                logging.info("Tracking with knowledge transfer")
                if not self.video.identity_transfer:
                    logging.info("Reinitializing fully connected layers")
                    self.identification_model.apply(fc_weights_reinit)
                else:
                    logging.info(
                        "Identity transfer. Not reinitializing the fully connected"
                        " layers."
                    )
            except RuntimeError as exc:
                logging.error(
                    f"Could not load model {self.accumulation_network_params} to"
                    " transfer knowledge, following without knowledge nor identity"
                    " transfer.\n"
                    f"Raised error: {exc}"
                )
                self.identification_model = LearnerClassification.create_model(
                    self.accumulation_network_params
                )
                self.identification_model.apply(weights_xavier_init)
        else:
            self.identification_model = LearnerClassification.create_model(
                self.accumulation_network_params
            )
            self.identification_model.apply(weights_xavier_init)

        logging.info("Setting the first global fragment for accumulation")
        first_global_fragment = max(
            self.list_of_global_fragments.global_fragments,
            key=lambda gf: gf.minimum_distance_travelled,
        )

        self.video.first_frame_first_global_fragment.append(
            first_global_fragment.first_frame_of_the_core
        )

        identify_first_global_fragment_for_accumulation(
            first_global_fragment,
            self.video,
            network_params=self.accumulation_network_params,
            identification_model=self.identification_model,
        )

        # Order global fragments by distance to the first global fragment for the accumulation
        self.list_of_global_fragments.order_by_distance_to_the_frame(
            first_global_fragment.first_frame_of_the_core
        )

        # Instantiate accumulation manager
        self.accumulation_manager = AccumulationManager(
            self.video.id_images_file_paths,
            self.video.n_animals,
            self.list_of_fragments,
            self.list_of_global_fragments,
        )

        # Selecting the first global fragment is considered as
        # the 0 accumulation step
        self.video.init_accumulation_statistics_attributes()
        self.accumulate()

    def accumulate(self):
        logging.info("Entering accumulation loop")
        if self.accumulation_manager.new_global_fragments_for_training:
            # Training and identification continues
            if (
                self.accumulation_manager.current_step == 1
                and self.video.accumulation_trial == 0
            ):
                # first training finished
                self.video.protocol1_timer.finish()
                self.video.protocol2_timer.start()

            # Training and identification step
            perform_one_accumulation_step(
                self.accumulation_manager,
                self.video,
                self.identification_model,
                self.accumulation_network_params,
            )
            # Re-enter the function for the next step of the accumulation
            self.accumulate()

        elif (
            not self.video.protocol2_timer.finished
            and self.accumulation_manager.ratio_accumulated_images
            > conf.THRESHOLD_EARLY_STOP_ACCUMULATION
        ):
            # Accumulation stop because protocol 1 is successful
            self.save_after_first_accumulation()
            self.video.protocol1_timer.finish()
            logging.info("Protocol 1 successful")
            assign_remaining_fragments(
                self.list_of_fragments,
                self.identification_model,
                self.accumulation_network_params,
                self.video.identify_timer,
            )

        elif not self.video.protocol3_pretraining_timer.finished:
            logging.info("No more new global fragments")
            self.save_after_first_accumulation()

            if (
                self.accumulation_manager.ratio_accumulated_images
                >= conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
            ):
                self.video.protocol2_timer.finish()
                logging.info("Protocol 2 successful")
                assign_remaining_fragments(
                    self.list_of_fragments,
                    self.identification_model,
                    self.accumulation_network_params,
                    self.video.identify_timer,
                )

            else:
                self.video.protocol1_timer.finish()
                self.video.protocol2_timer.finish(raise_if_not_started=False)
                logging.warning(
                    "[red]Protocol 2 failed, protocol 3 is going to start",
                    extra={"markup": True},
                )
                ask_about_protocol3(
                    self.video.protocol3_action, self.video.number_of_error_frames
                )
                self.pretrain()
                self.accumulate()

        elif (
            self.video.accumulation_trial
            < conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS
            and self.accumulation_manager.ratio_accumulated_images
            < conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
        ):
            logging.warning("Accumulation Protocol 3 failed. Opening parachute ...")
            if self.video.accumulation_trial == 0:
                self.video.protocol3_accumulation_timer.start()
            else:
                self.save_and_update_accumulation_parameters_in_parachute()
            self.video.accumulation_trial += 1
            self.accumulation_parachute_init(self.video.accumulation_trial)

            self.video.init_accumulation_statistics_attributes()
            self.accumulate()

        else:
            logging.info("Accumulation after protocol 3 has been successful")
            self.video.protocol3_accumulation_timer.finish()

            self.save_after_second_accumulation()
            assign_remaining_fragments(
                self.list_of_fragments,
                self.identification_model,
                self.accumulation_network_params,
                self.video.identify_timer,
            )

        # Whether to re-enter the function for the next accumulation step
        if self.accumulation_manager.new_global_fragments_for_training:
            self.accumulate()

    def save_after_first_accumulation(self):
        """Set flags and save data"""
        logging.info("Saving first accumulation parameters")

        # if not self.restoring_first_accumulation:
        self.video.ratio_accumulated_images = (
            self.accumulation_manager.ratio_accumulated_images
        )
        self.video.percentage_of_accumulated_images = [
            self.video.ratio_accumulated_images
        ]
        self.video.save()
        self.list_of_fragments.save(self.video.fragments_path)
        self.list_of_global_fragments.save(self.video.global_fragments_path)

    """ pretraining """

    def pretrain(self):
        self.video.protocol3_pretraining_timer.start()
        create_dir(self.video.pretraining_folder, remove_existing=True)

        pretrain_network_params = NetworkParams(
            n_classes=self.video.n_animals,
            architecture=conf.IDCNN_NETWORK_NAME,
            save_folder=self.video.pretraining_folder,
            model_name="identification_network",
            image_size=self.video.id_image_size,
            scopes_layers_to_optimize=conf.LAYERS_TO_OPTIMISE_PRETRAINING,
            optimizer="SGD",
            schedule=[30, 60],
            optim_args={"lr": conf.LEARNING_RATE_IDCNN_ACCUMULATION, "momentum": 0.9},
            epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
            knowledge_transfer_folder=self.video.knowledge_transfer_folder,
        )
        pretrain_network_params.save()

        # Initialize network
        if pretrain_network_params.knowledge_transfer_folder:
            self.identification_model = LearnerClassification.load_model(
                pretrain_network_params, knowledge_transfer=True
            )
            self.identification_model.apply(fc_weights_reinit)
        else:
            self.identification_model = LearnerClassification.create_model(
                pretrain_network_params
            )
            self.identification_model.apply(weights_xavier_init)

        self.list_of_fragments.reset(roll_back_to="fragmentation")
        self.list_of_global_fragments.sort_by_distance_travelled()

        pretraining_counter = -1
        ratio_of_pretrained_images = 0.0
        max_ratio_of_pretrained_images = conf.MAX_RATIO_OF_PRETRAINED_IMAGES
        while ratio_of_pretrained_images < max_ratio_of_pretrained_images:
            pretraining_counter += 1
            logging.info(
                "[bold]New pretraining iteration[/], using the #%s global fragment",
                pretraining_counter,
            )
            pretrain_global_fragment(
                self.video.n_animals,
                self.identification_model,
                pretrain_network_params,
                self.list_of_global_fragments.global_fragments[pretraining_counter],
                self.list_of_fragments.id_images_file_paths,
            )
            ratio_of_pretrained_images = (
                self.list_of_fragments.ratio_of_images_used_for_pretraining
            )

            logging.debug(
                f"{ratio_of_pretrained_images:.2%} of the images have been used during"
                f" pretraining (if higher than {max_ratio_of_pretrained_images:.2%} we"
                " stop pretraining)"
            )

        self.video.protocol3_pretraining_timer.finish()

    """ parachute """

    def accumulation_parachute_init(self, iteration_number: int):
        logging.debug("Accumulation_parachute_init")
        logging.info("Starting accumulation %i", iteration_number)

        # delete = not self.processes_to_restore.get("protocol3_accumulation")

        self.video.create_accumulation_folder(
            iteration_number=iteration_number, delete=True
        )
        self.video.accumulation_trial = iteration_number
        self.list_of_fragments.reset(roll_back_to="fragmentation")

        logging.info(
            "Setting #%d global fragment for accumulation", iteration_number - 1
        )

        self.list_of_global_fragments.sort_by_distance_travelled()
        try:
            first_global_fragment = self.list_of_global_fragments.global_fragments[
                iteration_number - 1
            ]
        except IndexError:
            first_global_fragment = None  # TODO what if this happens

        self.video.first_frame_first_global_fragment.append(
            first_global_fragment.first_frame_of_the_core
            if first_global_fragment is not None
            else None
        )

        if first_global_fragment is not None:
            identify_first_global_fragment_for_accumulation(
                first_global_fragment,
                self.video,
                (
                    LearnerClassification.load_model(self.accumulation_network_params)
                    if self.video.identity_transfer
                    else None
                ),
                self.accumulation_network_params,
            )

        # Sort global fragments by distance
        self.list_of_global_fragments.order_by_distance_to_the_frame(
            self.video.first_frame_first_global_fragment[iteration_number - 1]
        )
        logging.warning(
            "first_frame_first_global_fragment %s",
            self.video.first_frame_first_global_fragment,
        )
        logging.info(
            "We will restore the network from a previous pretraining: %s",
            self.video.pretraining_folder,
        )

        # Set saving folders
        self.accumulation_network_params.save_folder = self.video.accumulation_folder

        # Set restoring model_file
        self.accumulation_network_params.restore_folder = self.video.pretraining_folder

        # TODO: allow to train only the fully connected layers
        self.accumulation_network_params.scopes_layers_to_optimize = [
            "fully-connected1",
            "fully_connected_pre_softmax",
        ]
        logging.info("Initializing accumulation network")

        # Load pretrained network
        self.identification_model = LearnerClassification.load_model(
            self.accumulation_network_params
        )

        # Re-initialize fully-connected layers
        self.identification_model.apply(fc_weights_reinit)

        # Instantiate accumualtion manager
        self.accumulation_manager = AccumulationManager(
            self.video.id_images_file_paths,
            self.video.n_animals,
            self.list_of_fragments,
            self.list_of_global_fragments,
        )

        logging.info("Start accumulation")

    def save_and_update_accumulation_parameters_in_parachute(self):
        logging.warning(
            "self.accumulation_manager.ratio_accumulated_images %.4f",
            self.accumulation_manager.ratio_accumulated_images,
        )
        self.video.ratio_accumulated_images = (
            self.accumulation_manager.ratio_accumulated_images
        )
        self.video.percentage_of_accumulated_images.append(
            self.video.ratio_accumulated_images
        )
        self.list_of_fragments.save(
            self.video.accumulation_folder / "list_of_fragments.json"
        )

    def save_after_second_accumulation(self):
        logging.info("Saving second accumulation parameters")
        # Save accumulation parameters
        self.save_and_update_accumulation_parameters_in_parachute()

        # Choose best accumulation
        self.video.accumulation_trial = int(
            np.argmax(self.video.percentage_of_accumulated_images)
        )

        # Update ratio of accumulated images and  accumulation folder
        self.video.ratio_accumulated_images = (
            self.video.percentage_of_accumulated_images[self.video.accumulation_trial]
        )
        self.video.create_accumulation_folder()

        # Load light list of fragments with identities of the best accumulation
        self.list_of_fragments = ListOfFragments.load(
            self.video.auto_accumulation_folder / "list_of_fragments.json"
        )

        # Save objects
        self.list_of_fragments.save(self.video.fragments_path)
        self.list_of_global_fragments.save(self.video.global_fragments_path)

        # set restoring folder
        logging.info("Restoring networks to best second accumulation")
        self.accumulation_network_params.restore_folder = self.video.accumulation_folder

        # TODO: allow to train only the fully connected layers
        self.accumulation_network_params.scopes_layers_to_optimize = [
            "fully-connected1",
            "fully_connected_pre_softmax",
        ]
        logging.info("Initializing accumulation network")

        # Load pretrained network
        self.identification_model = LearnerClassification.load_model(
            self.accumulation_network_params
        )

        # # Re-initialize fully-connected layers
        # self.identification_model.apply(fc_weights_reinit)

        logging.info("Sending model and criterion to %s", DEVICE)
        cudnn.benchmark = True  # make it train faster
        self.identification_model.to(DEVICE)

        self.video.save()


def ask_about_protocol3(protocol3_action: str, n_error_frames: int) -> None:
    """Raises a CustomError if protocol3_action is abort or aks and user answers abortion"""
    logging.info("Protocol 3 action: %s", protocol3_action)

    if protocol3_action == "abort":
        raise CustomError(
            "Protocol 3 was going to start but PROTOCOL3_ACTION is set to 'abort'"
        )
    if protocol3_action == "continue":
        return

    if protocol3_action != "ask":
        raise ValueError(
            f'PROTOCOL3_ACTION "{protocol3_action}" not in ("ask", "abort", "continue")'
        )

    if n_error_frames > 0:
        logging.info(
            "Protocol 3 is a very time consuming algorithm and, in most cases, it"
            " can be avoided by redefining the segmentation parameters. As"
            " [red]there are %d frames with more blobs than animals[/red], we"
            " recommend you to abort the tracking session now and go back to the"
            " Segmentation app focusing on not having reflections, shades, etc."
            " detected as blobs. Check the following general recommendations:\n   "
            " - Define a region of interest to exclude undesired noise blobs\n    -"
            " Shrink the intensity (or background difference) thresholds\n    -"
            " Toggle the use of the background subtraction\n    - Shrink the blob's"
            " area thresholds",
            n_error_frames,
            extra={"markup": True},
        )
    else:
        logging.info(
            "Protocol 3 is a very time consuming algorithm and, in most cases, it"
            " can be avoided by redefining the segmentation parameters. As"
            " [bold]there are NOT frames with more blobs than animals[/bold], the"
            " video is unlikely to have non-animal blobs. Even so, you can choose"
            " to abort the tracking session and redefine the segmentation"
            " parameters (specially shrinking the intensity (or background"
            " difference) thresholds) or to continue with Protocol 3.",
            extra={"markup": True},
        )

    abort = None
    valid_answers = {"abort": True, "a": True, "continue": False, "c": False}
    while abort is None:
        answer_str = input(
            "What do you want to do now? Abort [A] or Continue [C]? "
        ).lower()
        if answer_str not in valid_answers:
            logging.warning("Invalid answer")
            continue
        abort = valid_answers[answer_str]
        logging.info("Answer --> Abort? %s", abort)
    if abort:
        raise CustomError(
            "This is not an actual error: protocol 3 was going to start"
            " but PROTOCOL3_ACTION is set to 'ask' and used aborted."
        )
    return



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging
from pathlib import Path

import torch
from torch.backends import cudnn
from torch.nn import CrossEntropyLoss, Module
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import GlobalFragment
from idtrackerai.network import (
    DEVICE,
    LearnerClassification,
    NetworkParams,
    fc_weights_reinit,
)
from idtrackerai.utils import conf

from .identity_dataset import get_training_data_loaders, split_data_train_and_validation
from .identity_network import StopTraining, TrainIdentification


def pretrain_global_fragment(
    number_of_animals: int,
    identification_model: Module,
    network_params: NetworkParams,
    pretraining_global_fragment: GlobalFragment,
    id_images_file_paths: list[Path],
):
    """Performs pretraining on a single global fragments"""

    images, labels = pretraining_global_fragment.get_images_and_labels(
        id_images_file_paths
    )

    train_data, val_data = split_data_train_and_validation(
        images, labels, validation_proportion=conf.VALIDATION_PROPORTION
    )
    logging.info(
        "Pretraining with %d images, %d for training and %d for validation",
        len(images),
        len(train_data["images"]),
        len(val_data["images"]),
    )

    # Set data loaders
    train_loader, val_loader = get_training_data_loaders(
        number_of_animals, train_data, val_data
    )

    criterion = CrossEntropyLoss(weight=torch.tensor(train_data["weights"]))

    # Re-initialize fully-connected layers
    identification_model.apply(fc_weights_reinit)

    logging.info("Sending model and criterion to %s", DEVICE)
    cudnn.benchmark = True  # make it train faster
    identification_model.to(DEVICE)
    criterion.to(DEVICE)

    logging.info(f"Setting {network_params.optimizer} optimizer")
    if network_params.optimizer == "Adam":
        optimizer = torch.optim.Adam(
            identification_model.parameters(), **network_params.optim_args
        )
    elif network_params.optimizer == "SGD":
        optimizer = torch.optim.SGD(
            identification_model.parameters(), **network_params.optim_args
        )
    else:
        raise AttributeError(network_params.optimizer)

    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    learner = LearnerClassification(
        identification_model, criterion, optimizer, scheduler
    )

    stop_training = StopTraining(network_params.n_classes)

    TrainIdentification(
        learner, train_loader, val_loader, network_params, stop_training
    )

    for fragment in pretraining_global_fragment.individual_fragments:
        fragment.used_for_pretraining = True



import logging
from pathlib import Path

import toml
from qtpy.QtCore import Qt, QTimer
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import (
    QCheckBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QSpinBox,
    QSplitter,
    QVBoxLayout,
    QWidget,
)

from idtrackerai import Video
from idtrackerai.utils import pprint_dict
from idtrackerai_GUI_tools import GUIBase, LabelRangeSlider, QHLine, VideoPlayer

from .segmentation_widgets import (
    BkgWidget,
    BlobInfoWidget,
    FrameAnalyzer,
    IntensityThresholds,
    OpenVideoWidget,
    ROIWidget,
    TrackingIntervalsWidget,
)


class SegmentationGUI(GUIBase):
    """The main class implementing the Segmentation app.
    From here, all widgets are coordinated and connected.

    Parameters
    ----------
    GUIBase : QMainWindow
        Base configuration for all idtracker.ai GUIs
    """

    def __init__(self, video: Video | None = None, signals: dict | None = None):
        super().__init__()

        self.setWindowTitle("Segmentation App")
        self.signals = signals or {}
        self.video = video or Video()
        self.documentation_url = (
            "https://idtracker.ai/en/latest/user_guide/segmentation_app.html"
        )

        self.open_widget = OpenVideoWidget(self, self.video.frames_per_episode)
        self.videoPlayer = VideoPlayer(self)
        self.frame_analyzer = FrameAnalyzer()
        self.blobInfo = BlobInfoWidget()
        self.bkg_widget = BkgWidget(
            self,
            self.video.number_of_frames_for_background,
            self.video.background_subtraction_stat,
        )
        self.ROI_Widget = ROIWidget(self)
        self.tracking_interval = TrackingIntervalsWidget(self)
        self.widgets_to_close.append(self.videoPlayer)

        self.intensity_thresholds = IntensityThresholds(self, min=0, max=255)
        self.area_thresholds = LabelRangeSlider(
            parent=self, min=1, max=60000, block_upper=False
        )

        self.save_parameters = QPushButton("Save parameters")
        self.save_parameters.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.save_parameters.setShortcut("Ctrl+S")

        self.check_segm = QCheckBox("Stop tracking if #blobs > #animals")
        self.track_wo_id = QCheckBox("Track without identities")

        res_reduct_row = QHBoxLayout()
        resreduct_label = QLabel("Resolution")
        res_reduct_row.addWidget(resreduct_label)
        self.resreduct = QSpinBox()
        self.resreduct.setMaximum(100)
        self.resreduct.setMinimum(10)
        self.resreduct.setSingleStep(10)
        self.resreduct.setSuffix("%")
        res_reduct_row.addWidget(self.resreduct)
        res_reduct_row.setAlignment(Qt.AlignmentFlag.AlignLeft)

        n_animals_row = QHBoxLayout()
        n_animals_label = QLabel("Number of animals")
        n_animals_row.addWidget(n_animals_label)
        self.n_animals = QSpinBox()
        self.n_animals.setMaximum(100)
        self.n_animals.setMinimum(0)
        n_animals_row.addWidget(self.n_animals)
        n_animals_row.setAlignment(Qt.AlignmentFlag.AlignLeft)

        area_row = QHBoxLayout()
        area_th_label = QLabel("Blob area\nthresholds")
        area_th_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        area_row.addWidget(area_th_label)
        area_row.addWidget(self.area_thresholds)

        session_row = QHBoxLayout()
        session_label = QLabel("Session")
        session_row.addWidget(session_label)
        self.session = QLineEdit()
        self.session.setPlaceholderText("Example: test, experiment_32A, ...")
        session_row.addWidget(self.session)
        session_row.addWidget(self.save_parameters)

        self.close_and_track_btn = QPushButton("Close window and track video")
        self.close_and_track_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Connecting widgets
        self.open_widget.path_clicked.connect(self.videoPlayer.setCurrentFrame)
        self.open_widget.new_video_paths.connect(self.new_video_paths)
        self.open_widget.new_episodes.connect(self.bkg_widget.set_new_video_paths)
        self.open_widget.video_paths_reordered.connect(
            self.videoPlayer.reorder_video_paths
        )
        self.resreduct.editingFinished.connect(self.resreduct.clearFocus)
        self.resreduct.valueChanged.connect(
            lambda x: self.ROI_Widget.set_resolution_reduction(x / 100)
        )
        self.resreduct.valueChanged.connect(
            lambda x: self.videoPlayer.set_resolution_reduction(x / 100)
        )
        self.resreduct.valueChanged.connect(
            lambda x: self.frame_analyzer.set_resolution_reduction(x / 100)
        )
        self.n_animals.editingFinished.connect(self.n_animals.clearFocus)
        self.n_animals.valueChanged.connect(self.blobInfo.setNAnimals)
        self.tracking_interval.newValue.connect(self.open_widget.set_tracking_interval)
        self.tracking_interval.newValue.connect(self.blobInfo.setTrackingIntervals)
        self.intensity_thresholds.newValue.connect(
            self.frame_analyzer.set_intensity_ths
        )
        self.session.editingFinished.connect(self.session.clearFocus)
        self.save_parameters.clicked.connect(self.save_parameters_func)
        self.area_thresholds.valueChanged.connect(self.frame_analyzer.set_area_ths)
        self.close_and_track_btn.clicked.connect(self.close_and_track_video)
        self.ROI_Widget.valueChanged.connect(self.frame_analyzer.set_ROI_mask)
        self.ROI_Widget.needToDraw.connect(self.videoPlayer.update)
        self.bkg_widget.new_bkg_data.connect(self.frame_analyzer.set_bkg)
        self.bkg_widget.new_bkg_data.connect(self.intensity_thresholds.bkg_changed)
        self.frame_analyzer.new_areas.connect(self.blobInfo.setAreas)
        self.frame_analyzer.new_parameters.connect(self.videoPlayer.update)
        self.videoPlayer.painting_time.connect(self.frame_analyzer.paint_on_canvas)
        self.videoPlayer.painting_time.connect(self.ROI_Widget.paint_on_canvas)
        self.videoPlayer.canvas.click_event.connect(self.ROI_Widget.click_event)

        # Tooltips texts
        tooltips = toml.load(Path(__file__).parent / "tooltips.toml")

        self.open_widget.button_open.setToolTip(tooltips["open_btn"])
        self.open_widget.single_file_label.setToolTip(tooltips["open_path_label"])
        self.open_widget.list_of_files.setToolTip(tooltips["open_path_list"])
        self.tracking_interval.setToolTip(tooltips["tacking_interval"])
        self.ROI_Widget.setToolTip(tooltips["region_of_interest"])
        self.bkg_widget.setToolTip(tooltips["background_subtraction"])
        self.n_animals.setToolTip(tooltips["number_of_animals"])
        n_animals_label.setToolTip(tooltips["number_of_animals"])
        self.check_segm.setToolTip(tooltips["check_segm"])
        self.area_thresholds.setToolTip(tooltips["area_thresholds"])
        area_th_label.setToolTip(tooltips["area_thresholds"])
        self.resreduct.setToolTip(tooltips["resolution_reduction"])
        resreduct_label.setToolTip(tooltips["resolution_reduction"])
        self.track_wo_id.setToolTip(tooltips["track_wo_id"])
        self.save_parameters.setToolTip(tooltips["save_params"])
        self.close_and_track_btn.setToolTip(tooltips["close_and_track"])
        self.blobInfo.setToolTip(tooltips["blobs_info"])
        self.session.setToolTip(tooltips["session_name"])
        session_label.setToolTip(tooltips["session_name"])
        self.intensity_thresholds.setToolTips(
            tooltips["intensity_thresholds_nobkg"],
            tooltips["intensity_thresholds_yesbkg"],
        )

        # Define widget structure
        left_layout = QVBoxLayout()

        widgets = (
            QHLine(),
            res_reduct_row,
            self.tracking_interval,
            self.ROI_Widget,
            QHLine(),
            n_animals_row,
            self.bkg_widget,
            self.intensity_thresholds,
            area_row,
            QHLine(),
            self.check_segm,
            self.track_wo_id,
        )
        self.open_widget.layout().setContentsMargins(0, 8, 0, 0)
        left_layout.addWidget(self.open_widget)
        for widget in widgets:
            if isinstance(widget, (QVBoxLayout, QHBoxLayout)):
                widget.setContentsMargins(0, 0, 0, 0)
                superwidget = QWidget()
                superwidget.setLayout(widget)
                left_layout.addWidget(
                    superwidget, alignment=Qt.AlignmentFlag.AlignVCenter
                )
            else:
                lay = widget.layout()
                if lay is not None:
                    lay.setContentsMargins(0, 0, 0, 0)
                left_layout.addWidget(
                    widget,
                    alignment=Qt.AlignmentFlag.AlignVCenter,
                    stretch=1 if isinstance(widget, QHLine) else 3,
                )
        left_layout.addLayout(session_row)
        left_layout.addWidget(self.close_and_track_btn)

        self.right_splitter = QSplitter(Qt.Orientation.Vertical)
        self.right_splitter.addWidget(self.blobInfo)
        self.videoPlayer.layout().setContentsMargins(8, 8, 0, 0)
        self.right_splitter.addWidget(self.videoPlayer)
        self.right_splitter.setSizes([200, 600])

        left = QWidget()
        left.setLayout(left_layout)

        main_splitter = QSplitter(Qt.Orientation.Horizontal)
        main_splitter.addWidget(left)
        main_splitter.addWidget(self.right_splitter)
        main_splitter.setSizes([400, 600])
        self.centralWidget().layout().addWidget(main_splitter)
        self.list_of_widgets = self.get_list_of_widgets(left_layout)
        for widget in self.list_of_widgets:
            widget.setEnabled(False)
        self.right_splitter.setEnabled(False)
        self.enabled = False
        self.open_widget.setEnabled(True)

        self.setTabOrder(self.resreduct, self.videoPlayer.canvas)
        self.setTabOrder(self.videoPlayer.canvas, self.resreduct)
        for widget in self.findChildren(QCheckBox):
            assert isinstance(widget, QWidget)
            widget.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        QTimer.singleShot(0, self.load_parameters)

    def load_parameters(self):
        """Loads configuration from `load_dict` setting the corresponding widgets status

        Parameters
        ----------
        load_dict : dict
            Parameters to load
        """
        self.open_widget.open_video_paths(self.video.video_paths)
        self.resreduct.setValue(int(self.video.resolution_reduction * 100))
        self.tracking_interval.setValue(self.video.tracking_intervals)
        self.ROI_Widget.setValue(self.video.roi_list)
        self.intensity_thresholds.setValue(self.video.intensity_ths or (0, 130))
        self.area_thresholds.setValue(self.video.area_ths or (50, 10000))
        self.n_animals.setValue(self.video.number_of_animals)
        self.track_wo_id.setChecked(self.video.track_wo_identities)
        self.check_segm.setChecked(self.video.check_segmentation)
        self.session.setText(self.video.session)
        self.bkg_widget.checkBox.setChecked(self.video.use_bkg)
        self.videoPlayer.update()

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self.ROI_Widget.add.setChecked(False)

    def close_and_track_video(self):
        """Action when clicked "close and track video".
        It gathers widgets parameters, writes them in self.user_params and exits"""
        parameters = self.out_parameters()
        if self.unacceptable_parameters(parameters):
            return

        logging.info(pprint_dict(parameters, "GUI params"), extra={"markup": True})
        self.video.set_parameters(**parameters)
        self.video.background_from_segmentation_gui = self.bkg_widget.getBkg()
        # signal to start tracking after closing app
        self.signals["run_idtrackerai"] = True
        self.close()

    def getSessionName(self) -> str:
        return self.session.text() or "no_name"

    def out_parameters(self) -> dict:
        """Generates dict of all widgets content

        Returns
        -------
        dict
            Parameter dict containing all widgets content
        """
        return {
            "session": self.getSessionName(),
            "video_paths": self.open_widget.getVideoPaths(),
            "intensity_ths": self.intensity_thresholds.value(),
            "area_ths": self.area_thresholds.value(),
            "tracking_intervals": self.tracking_interval.value(),
            "number_of_animals": self.n_animals.value(),
            "use_bkg": self.bkg_widget.checkBox.isChecked(),
            "check_segmentation": self.check_segm.isChecked(),
            "resolution_reduction": self.resreduct.value() / 100,
            "track_wo_identities": self.track_wo_id.isChecked(),
            "roi_list": self.ROI_Widget.getValue(),
        }

    def unacceptable_parameters(self, parameters: dict) -> bool:
        if (
            parameters["number_of_animals"] == 0
            and not parameters["track_wo_identities"]
        ):
            QMessageBox.warning(
                self,
                "Missing parameters",
                'Please, define the number of animals in the video or check "Track'
                ' without identities".\n\nEven if tracking without identities,'
                " adding the number of animals is recommended to improve the"
                " individual/crossing blob detection.",
            )
            return True
        if parameters["roi_list"] is not None and len(parameters["roi_list"]) == 0:
            QMessageBox.warning(
                self,
                "Missing parameters",
                "Please, add a region of interest or uncheck the Regions of"
                " interest parameter.",
            )
            return True
        return False

    def save_parameters_func(self):
        parameters = self.out_parameters()

        if self.unacceptable_parameters(parameters):
            return

        fileName, _ = QFileDialog.getSaveFileName(
            self,
            "Save parameter file",
            str(Path.cwd() / (self.getSessionName() + ".toml")),
            filter="TOML (*.toml)",
        )
        if not fileName:
            return

        with open(fileName, "w", encoding="utf_8") as file:
            for key, value in parameters.items():
                file.write(f"{key} = {toml_format(value)}\n")

    def new_video_paths(
        self,
        video_paths: list[str],
        video_size: tuple[int, int],
        n_frames: int,
        fps: int,
        episodes: list,
    ):
        self.ROI_Widget.set_video_size(video_size)
        self.videoPlayer.setEnabled(False)
        self.tracking_interval.reset(n_frames)
        self.frame_analyzer.drawn_frame = -1
        self.bkg_widget.set_new_video_paths(video_paths, episodes)
        self.ROI_Widget.list.ListChanged.emit()
        self.videoPlayer.update_video_paths(video_paths, n_frames, video_size, fps)

        if not self.enabled:
            for widget in self.list_of_widgets:
                widget.setEnabled(True)
            self.enabled = True
            self.right_splitter.setEnabled(True)

        self.videoPlayer.setEnabled(True)
        self.videoPlayer.update()


def toml_format(value, width=50) -> str:
    """Custom .toml formatter.

    Parameters
    ----------
    value : Any
        The value to format
    width : int, optional
        Maximum line width before wrapping, by default 50

    Returns
    -------
    str
        Formatted value
    """
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float, str)):
        return repr(value)
    if value is None:
        return '""'
    if not value:
        return "[]"
    value = list(value)

    if len(repr(value)) < width:
        return repr(value)

    s = "[\n"
    for item in value:
        s += f"    {repr(item)},\n"
    s += "]"
    return s



import cv2
import numpy as np
from qtpy.QtCore import Signal
from qtpy.QtGui import QColor, QPolygon
from qtpy.QtWidgets import QWidget

from idtrackerai.animals_detection.segmentation import process_frame
from idtrackerai_GUI_tools import CanvasPainter


class FrameAnalyzer(QWidget):
    new_areas = Signal(int, list)
    new_parameters = Signal()

    def set_bkg(self, bkg_model):
        self.bkg_model = bkg_model
        self.bkg_model_resreduct = bkg_model
        self.use_bkg = bkg_model is not None

        if bkg_model is not None and self.resolution_reduction != 1:
            self.bkg_model_resreduct = cv2.resize(
                self.bkg_model,
                None,  # type: ignore
                fx=self.resolution_reduction,
                fy=self.resolution_reduction,
                interpolation=cv2.INTER_AREA,
            )

        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_ROI_mask(self, ROI_mask: np.ndarray | None):
        self.ROI_mask = ROI_mask
        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_resolution_reduction(self, resolution_reduction: float):
        self.resolution_reduction = resolution_reduction

        if resolution_reduction != 1:
            if self.bkg_model is not None:
                self.bkg_model_resreduct = cv2.resize(
                    self.bkg_model,
                    None,  # type: ignore
                    fx=resolution_reduction,
                    fy=resolution_reduction,
                    interpolation=cv2.INTER_AREA,
                )
        else:
            self.bkg_model_resreduct = self.bkg_model

        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_intensity_ths(self, intensity_ths: list[int]):
        self.intensity_ths = intensity_ths
        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_area_ths(self, area_ths: list[int]):
        self.area_ths = area_ths
        self.need_to_redraw = True
        self.new_parameters.emit()

    def __init__(self):
        super().__init__()

        self.use_bkg = False
        self.bkg_model = None
        self.bkg_model_resreduct = None
        self.ROI_mask = None
        self.intensity_ths = [0, 1]
        self.area_ths = [0, 1]
        self.resolution_reduction = 1
        self.blob_polygons: list[QPolygon] = []
        self.drawn_frame = -1

    def process_frame(self, frame):
        self.areas, contours, gray_frame = process_frame(
            frame,
            bkg_model=self.bkg_model_resreduct,
            ROI_mask=self.ROI_mask,
            resolution_reduction=self.resolution_reduction,
            intensity_ths=self.intensity_ths,
            area_ths=self.area_ths,
        )

        self.n_blobs = len(contours)
        for i, contour in enumerate(contours):
            if i == len(self.blob_polygons):
                self.blob_polygons.append(QPolygon())
            self.blob_polygons[i].setPoints(*contour.ravel())

    def paint_on_canvas(self, painter: CanvasPainter, frame_number: int, frame):
        if self.drawn_frame != frame_number or self.need_to_redraw:
            self.process_frame(frame)
            self.new_areas.emit(frame_number, self.areas)
            self.need_to_redraw = False
        painter.setBrush(QColor(60, 160, 255, 150))
        painter.setPenColor(QColor(0x286384))
        for i in range(self.n_blobs):
            painter.drawPolygon(self.blob_polygons[i])
        self.drawn_frame = frame_number



"""isort:skip_file"""

# NetworkParams should be loaded before LearnerClassification
from .utils import normalize, fc_weights_reinit, weights_xavier_init, DEVICE
from .network_params import NetworkParams
from .learners import LearnerClassification
from .train import train, evaluate


__all__ = [
    "evaluate",
    "LearnerClassification",
    "train",
    "weights_xavier_init",
    "normalize",
    "fc_weights_reinit",
    "NetworkParams",
    "DEVICE",
]



from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtWidgets import QHBoxLayout, QLabel, QWidget

from idtrackerai_GUI_tools import LabeledSlider, LabelRangeSlider


class IntensityThresholds(QWidget):
    newValue = Signal(object)

    def __init__(self, parent, min, max):
        super().__init__()
        self.parent_widget = parent
        self.label_nobkg = QLabel("Blob intensity\nthresholds")
        self.label_nobkg.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label_yesbkg = QLabel("Background\ndifference threshold")
        self.label_yesbkg.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label_yesbkg.setVisible(False)
        self.range_slider = LabelRangeSlider(parent=parent, min=min, max=max)
        self.simple_slider = LabeledSlider(parent, min=min, max=max)
        self.simple_slider.setVisible(False)

        self.range_slider.valueChanged.connect(self.newValue.emit)
        self.simple_slider.valueChanged.connect(lambda x: self.newValue.emit((0, x)))
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.label_nobkg)
        layout.addWidget(self.label_yesbkg)
        layout.addWidget(self.range_slider)
        layout.addWidget(self.simple_slider)

    def bkg_changed(self, bkg):
        if bkg is None:
            self.label_nobkg.setVisible(True)
            self.label_yesbkg.setVisible(False)
            self.range_slider.setVisible(True)
            self.simple_slider.setVisible(False)
            self.newValue.emit(self.range_slider.value())
        else:
            self.label_nobkg.setVisible(False)
            self.label_yesbkg.setVisible(True)
            self.range_slider.setVisible(False)
            self.simple_slider.setVisible(True)
            self.newValue.emit((0, self.simple_slider.value()))

    def setValue(self, value):
        self.range_slider.setValue(value)
        self.simple_slider.setValue(value[1])

    def value(self):
        if self.range_slider.isVisible():
            return self.range_slider.value()
        return 0, self.simple_slider.value()

    def setToolTips(self, tooltip_nobkg: str, tooltip_yesbkg: str):
        self.label_nobkg.setToolTip(tooltip_nobkg)
        self.label_yesbkg.setToolTip(tooltip_yesbkg)
        self.range_slider.setToolTip(tooltip_nobkg)
        self.simple_slider.setToolTip(tooltip_yesbkg)



import numpy as np
from qtpy.QtCore import QEvent, QPointF, Qt, Signal  # type: ignore
from qtpy.QtGui import QColorConstants, QKeyEvent
from qtpy.QtWidgets import (
    QComboBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QToolButton,
    QVBoxLayout,
)
from scipy.interpolate import interp1d

from idtrackerai import ListOfBlobs
from idtrackerai_GUI_tools import (
    CanvasMouseEvent,
    CanvasPainter,
    WrappedLabel,
    key_event_modifier,
)


class CustomComboBox(QComboBox):
    def keyPressEvent(self, e: QKeyEvent):
        event = key_event_modifier(e)
        if event is not None:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, e: QKeyEvent):
        event = key_event_modifier(e)
        if event is not None:
            super().keyReleaseEvent(event)


class Interpolator(QGroupBox):
    interpolation_kinds = {"linear": 1, "quadratic": 2, "cubic": 3, "5th order": 5}
    neew_to_draw = Signal()
    update_trajectories = Signal(int, int)
    go_to_frame = Signal(int)
    preload_frames = Signal(int, int)
    interpolation_accepted = Signal()
    enabled_changed = Signal(bool)

    def __init__(self) -> None:
        super().__init__()
        layout = QVBoxLayout()
        self.setLayout(layout)

        self.warning = WrappedLabel()
        self.warning.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self.warning)
        self.warning.setVisible(False)

        self.goto_btn = QPushButton()
        layout.addWidget(self.goto_btn)
        self.goto_btn.setVisible(False)

        self.info_label = WrappedLabel()
        layout.addWidget(self.info_label)

        range_row = QHBoxLayout()
        range_row.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.start_btn = QToolButton()
        self.end_btn = QToolButton()
        self.start_btn.clicked.connect(lambda: self.go_to_frame.emit(self.start - 1))
        self.end_btn.clicked.connect(lambda: self.go_to_frame.emit(self.end))
        range_row.addWidget(QLabel("From"))
        range_row.addWidget(self.start_btn)
        range_row.addWidget(QLabel("to"))
        range_row.addWidget(self.end_btn)
        layout.addLayout(range_row)

        self.interpolation_order_box = CustomComboBox()
        self.interpolation_order_box.addItems(self.interpolation_kinds.keys())
        self.interpolation_order_box.setCurrentText("cubic")
        self.interpolation_order_box.currentTextChanged.connect(self.new_interp_type)
        order_row = QHBoxLayout()
        self.interpolation_order_label = WrappedLabel("Interpolation order")
        order_row.addWidget(self.interpolation_order_label)
        order_row.addWidget(self.interpolation_order_box)
        layout.addLayout(order_row)

        self.input_size_row = QHBoxLayout()
        self.input_size_row.addWidget(WrappedLabel("Input size"))
        for value in (10, 150, 1500):
            btn = QRadioButton(str(value))
            if value == 10:
                btn.setChecked(True)
            btn.clicked.connect(self.new_input_size)
            self.input_size_row.addWidget(btn)
        layout.addLayout(self.input_size_row)
        self.input_size = 10

        remove_centroid = QPushButton("Remove centroid [R]")
        remove_centroid.setShortcut(Qt.Key.Key_R)
        remove_centroid.clicked.connect(self.remove_current_centroid)
        layout.addWidget(remove_centroid)

        apply_row = QHBoxLayout()
        style = self.style()
        self.abort_btn = QPushButton(
            style.standardIcon(style.StandardPixmap.SP_DialogCancelButton),
            "Abort [Esc]",
        )
        self.apply_btn = QPushButton(
            style.standardIcon(style.StandardPixmap.SP_DialogOkButton), "Apply [Ctrl+A]"
        )
        self.abort_btn.setShortcut(Qt.Key.Key_Escape)
        self.apply_btn.setShortcut("Ctrl+A")
        self.abort_btn.clicked.connect(lambda: self.setActivated(False))
        self.apply_btn.clicked.connect(self.apply_interpolation)
        apply_row.addWidget(self.abort_btn)
        apply_row.addWidget(self.apply_btn)
        layout.addLayout(apply_row)

        self.setActivated(False)
        self.animal_id: int = -1
        self.interp1d: interp1d
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    def trajectories_have_been_updated(self):
        if self.isEnabled():
            self.build_interpolator()

    def changeEvent(self, event: QEvent):
        if event.type() == QEvent.Type.EnabledChange:
            self.enabled_changed.emit(self.isEnabled())

    def new_interp_type(self, kind: str):
        self.interp1d = interp1d(
            self.interp1d.x,
            self.interp1d.y,
            kind=self.interpolation_kinds[kind],  # type: ignore
            fill_value="extrapolate",  # type: ignore
            assume_sorted=True,
        )
        self.neew_to_draw.emit()

    def new_input_size(self):
        btn = self.sender()
        assert isinstance(btn, QRadioButton)
        self.input_size = int(btn.text())
        self.build_interpolator()

    def set_interpolation_params(self, animal_id, start, end):
        self.start = start
        self.end = end
        self.animal_id = animal_id - 1
        self.expand_start()
        self.expand_end()
        self.preload_frames.emit(max(0, self.start - 10), self.start)
        self.build_interpolator()

    def build_interpolator(self):
        self.interpolation_range = range(self.start, self.end)
        self.continuous_interpolation_range = np.arange(
            max(self.start - 1, 0), self.end + 0.1, 0.2
        )
        self.entire_range = range(
            max(0, self.start - self.input_size),
            min(self.n_frames, self.end + self.input_size),
        )

        n_duplicated = np.count_nonzero(
            self.duplicated[self.entire_range, self.animal_id]
        )
        if n_duplicated:
            first_duplicated = (
                self.duplicated[self.entire_range, self.animal_id].argmax()
                + self.entire_range.start
            )
            self.warning.setText(
                f'<font color="red">There are {n_duplicated} frames where identity'
                f" {self.animal_id+1} appears duplicated. It is highly recommended to"
                " solve this before proceeding. The first duplication appears at frame"
                f" {first_duplicated}"
            )
            self.goto_btn.setText(f"Go to {first_duplicated}")
            self.goto_btn.clicked.connect(
                lambda: self.go_to_frame.emit(first_duplicated)
            )
            self.warning.setVisible(True)
            self.goto_btn.setVisible(True)
        else:
            self.warning.setVisible(False)
            self.goto_btn.setVisible(False)

        times_were_not_nan = np.asarray(self.entire_range)[
            ~np.isnan(self.trajectories[self.entire_range, self.animal_id, 0])
        ]

        self.interp1d = interp1d(
            times_were_not_nan,
            self.trajectories[times_were_not_nan, self.animal_id].T,
            kind=self.interpolation_kinds[
                self.interpolation_order_box.currentText()
            ],  # type:ignore
            fill_value="extrapolate",  # type:ignore
            assume_sorted=True,
        )
        self.info_label.setText(
            f'Interpolating identity <span style="font-weight:600">{self.animal_id+1}'
        )
        self.setActivated(True)
        self.setFocus()

    def remove_current_centroid(self):
        if self.current_frame not in self.entire_range:
            QMessageBox.warning(
                self,
                "Interpolator message",
                "Cannot remove current centroid outside interpolation "
                f"range ({self.entire_range.start} -> {self.entire_range.stop})",
            )
            return

        centroid_to_remove = self.trajectories[self.current_frame, self.animal_id]
        if np.isnan(centroid_to_remove[0]):
            QMessageBox.warning(
                self,
                "Interpolator message",
                "Cannot remove current centroid because it does not exist",
            )
            return

        self.list_of_blobs.remove_centroid(
            self.current_frame, centroid_to_remove, self.animal_id + 1
        )
        self.update_trajectories.emit(self.current_frame, self.current_frame + 1)

        if self.current_frame == self.start - 1:
            self.expand_start()
            self.go_to_frame.emit(self.start - 1)
        elif self.current_frame == self.end:
            self.expand_end()
            self.go_to_frame.emit(self.end)

        self.build_interpolator()

    def expand_start(self):
        for frame in range(self.start - 1, -1, -1):
            if not np.isnan(self.trajectories[frame, self.animal_id, 0]):
                self.start = frame + 1
                break

    def expand_end(self):
        for frame in range(self.end, self.n_frames):
            if not np.isnan(self.trajectories[frame, self.animal_id, 0]):
                self.end = frame
                break

    def click_event(self, event: CanvasMouseEvent):
        if (
            event.button != Qt.MouseButton.RightButton
            or not self.isEnabled()
            or self.current_frame not in self.interpolation_range
        ):
            return

        current_postion = self.trajectories[self.current_frame, self.animal_id]
        already_has_a_centroid = not np.isnan(current_postion[0])
        if already_has_a_centroid:
            self.list_of_blobs.update_centroid(
                self.current_frame, self.animal_id + 1, current_postion, event.xy_data
            )
        else:
            self.list_of_blobs.add_centroid(
                self.current_frame, self.animal_id + 1, event.xy_data
            )
        self.update_trajectories.emit(self.current_frame, self.current_frame + 1)

    def setActivated(self, activated: bool):
        self.setEnabled(activated)
        if not activated:
            self.warning.setVisible(False)
            self.goto_btn.setVisible(False)
            self.info_label.setText(
                'Select some errors of kind "Miss id" of '
                '"Jump" to start an interpolation process'
            )
        self.neew_to_draw.emit()

    def apply_interpolation(self):
        for new_centroid, frame in zip(
            self.interp1d(self.interpolation_range).T, self.interpolation_range
        ):
            if np.isnan(self.trajectories[frame, self.animal_id, 0]):
                self.list_of_blobs.add_centroid(frame, self.animal_id + 1, new_centroid)
        self.setEnabled(False)
        self.interpolation_accepted.emit()
        self.update_trajectories.emit(self.start, self.end)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = value
        self.start_btn.setText(f"frame {value-1}")
        self.start_btn.setToolTip(f"Go to frame {value-1}")

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        self._end = value
        self.end_btn.setText(f"frame {value}")
        self.end_btn.setToolTip(f"Go to frame {value}")

    def set_references(
        self,
        traj: np.ndarray,
        unidentified: np.ndarray,
        duplicated: np.ndarray,
        list_of_blobs: ListOfBlobs,
    ):
        self.list_of_blobs = list_of_blobs
        self.trajectories = traj
        self.unidentified = unidentified
        self.duplicated = duplicated
        self.n_frames = self.trajectories.shape[0]

    def paint_on_canvas(self, painter: CanvasPainter, frame: int):
        self.current_frame = frame
        x_input = self.interp1d.x
        y_input = self.interp1d.y.T

        # interpolated points
        painter.setPenColor(QColorConstants.White)
        painter.setBrush(QColorConstants.White)
        for point in self.interp1d(self.interpolation_range).T:
            painter.drawBigPoint(*point)

        # continuum interpolated range
        painter.drawPolyline(
            [
                QPointF(*xy)
                for xy in self.interp1d(self.continuous_interpolation_range).T
            ]  # type: ignore
        )

        # interpolator input data
        painter.setPenColor(QColorConstants.Red)
        painter.setBrush(QColorConstants.Red)
        painter.drawPolyline([QPointF(*xy) for xy in y_input[x_input < self.start]])  # type: ignore
        painter.drawPolyline([QPointF(*xy) for xy in y_input[x_input >= self.end]])  # type: ignore
        for point in y_input:
            painter.drawBigPoint(*point)

        # actual point
        if (
            self.current_frame in self.interp1d.x
            or self.current_frame in self.interpolation_range
        ):
            painter.setPenColor(QColorConstants.White)
            painter.drawBigPoint(*self.interp1d(frame))



import ast
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

from idtrackerai import Video
from idtrackerai.utils import CustomError


def Bool(value: str):
    valid = {"true": True, "t": True, "1": True, "false": False, "f": False, "0": False}

    lower_value = value.lower()
    if lower_value not in valid:
        raise ValueError
    return valid[lower_value]


def path(value: str):
    return_path = Path(value).expanduser().resolve().absolute()
    if not return_path.exists():
        raise CustomError(f'The path "{return_path}" does not exist.')
    return return_path


def pair_of_ints(value: str):
    out = ast.literal_eval(value)
    if not isinstance(out, (tuple, list)):
        raise ValueError

    out = list(out)
    if len(out) != 2:
        raise ValueError
    if any(not isinstance(x, int) for x in out):
        raise ValueError
    return out


def get_parser(defaults: Optional[dict] = None) -> ArgumentParser:
    defaults = defaults or {}

    parser = ArgumentParser(
        prog="idtracker.ai", epilog="For more info visit https://idtracker.ai"
    )

    def add_argument(name: str, help: str, type, metavar: str = "", **kwargs):
        name = name.lower()

        metavar = f"<{type.__name__.lower()}>"

        if "choices" in kwargs:
            help += f' (choices: {", ".join(kwargs["choices"])})'

        if name.upper() in defaults:
            help += f" (default: {defaults[name.upper()]})"
        elif name.lower() in defaults:
            help += f" (default: {defaults[name.lower()]})"

        parser.add_argument(
            "--" + name, help=help + ".", type=type, metavar=metavar, **kwargs
        )

    add_argument(
        "load",
        help="Primary .toml file to load session parameters",
        type=path,
        metavar="session_parameters",
        dest="session_parameters",
    )

    add_argument(
        "settings",
        help="Secondary .toml file to load general settings",
        type=path,
        metavar="general_settings",
        dest="general_settings",
    )

    parser.add_argument(
        "--track", help="Track the video without launching the GUI", action="store_true"
    )

    add_argument(
        "tracking_intervals",
        help=(
            "Tracking intervals in frames. "
            'Examples: "0,100", "[0,100]", "[0,100] [150,200] ...". '
            "If none, the whole video is tracked"
        ),
        type=pair_of_ints,
        nargs="+",
    )
    add_argument(
        "identity_transfer",
        help="If true, identities from knowledge transfer folder are transferred",
        type=Bool,
    )
    add_argument(
        "intensity_ths",
        help=(
            "Blob's intensity thresholds. When using background subtraction, the"
            " background difference threshold is the second value of these intensity"
            " thresholds"
        ),
        type=int,
        nargs=2,
    )
    add_argument("area_ths", help="Blob's areas thresholds", type=int, nargs=2)
    add_argument(
        "number_of_animals",
        help="Number of different animals that appear in the video",
        type=int,
    )
    add_argument(
        "output_dir",
        help=(
            "Output directory where session folder will be saved to, default is video"
            " paths parent directory"
        ),
        type=path,
    )
    add_argument(
        "resolution_reduction", help="Video resolution reduction ratio", type=float
    )
    add_argument(
        "check_segmentation",
        help="Check all frames have less or equal number of blobs than animals",
        type=Bool,
    )
    add_argument(
        "ROI_list",
        help="List of polygons defining the Region Of Interest",
        type=str,
        nargs="+",
    )
    add_argument(
        "use_bkg",
        help="Compute and extract background to improve blob identification",
        type=Bool,
    )
    add_argument(
        "video_paths",
        help="List of paths to the video files to track",
        type=path,
        nargs="+",
    )
    add_argument("session", help="Name of the session", type=str)
    add_argument(
        "track_wo_identities",
        "Track the video ignoring identities (without AI)",
        type=Bool,
    )
    add_argument(
        "CONVERT_TRAJECTORIES_TO_CSV_AND_JSON",
        "If true, trajectories files are gonna be copied to .csv and .json files",
        type=Bool,
    )
    add_argument(
        "ADD_TIME_COLUMN_TO_CSV",
        "If true, adds a time column (in seconds) to csv trajectory files",
        type=Bool,
    )
    add_argument(
        "FRAMES_PER_EPISODE",
        "Maximum number of frames for each video episode (used to parallelize some"
        " processes)",
        type=int,
    )
    add_argument(
        "KNOWLEDGE_TRANSFER_FOLDER",
        "Path to the session to transfer knowledge from",
        type=path,
    )
    add_argument(
        "BACKGROUND_SUBTRACTION_STAT",
        "Statistical method to compute the background",
        type=str,
        choices=["median", "mean", "max", "min"],
    )
    add_argument(
        "protocol3_action",
        "Choose what to do when protocol 1 and 2 fail and protocol 3 is going to start",
        type=str,
        choices=["ask", "abort", "continue"],
    )
    add_argument(
        "NUMBER_OF_FRAMES_FOR_BACKGROUND",
        "Number of frames used to compute the background",
        type=int,
    )
    add_argument(
        "number_of_parallel_workers",
        "Maximum number of jobs to parallelize segmentation and "
        "identification image creation. A negative value means using the number "
        "of CPUs in the system minus the specified value. Zero means using half "
        "of the number of CPUs in the system",
        type=int,
    )
    add_argument(
        "DATA_POLICY",
        "Type of data policy indicating the data in the session folder not to be"
        "erased when successfully finished a tracking",
        choices=[
            "trajectories",
            "validation",
            "knowledge_transfer",
            "idmatcher.ai",
            "all",
        ],
        type=str,
    )
    add_argument(
        "ID_IMAGE_SIZE",
        "The size of the identification images used in the tracking",
        type=int,
    )
    return parser


def get_argparser_help():
    """Used to display argument options in docs

    Returns
    -------
    str
        idtracker.ai argument parser help
    """
    return get_parser(Video.__dict__).format_help()


def parse_args(defaults: dict | None = None):
    parser = get_parser(defaults or Video.__dict__)
    return {k: v for k, v in vars(parser.parse_args()).items() if v is not None}


if __name__ == "__main__":
    print(get_argparser_help())



import ast

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QCheckBox, QHBoxLayout, QLineEdit, QMessageBox, QWidget

from idtrackerai_GUI_tools import LabelRangeSlider


class TrackingIntervalsWidget(QWidget):
    newValue = Signal(object)

    def __init__(self, parent):
        super().__init__()
        self.checkbox = QCheckBox("Tracking interval")
        self.checkbox.clicked.connect(self.checkbox_clicked)
        self.range_slider = LabelRangeSlider(parent=parent, min=0, max=1)

        self.range_slider.setVisible(False)

        self.multiple_CheckBox = QCheckBox("Multiple")
        self.multiple_CheckBox.setVisible(False)
        self.range_slider.valueChanged.connect(self.emit)
        self.checkbox.clicked.connect(self.emit)

        self.multiple_CheckBox.stateChanged.connect(self.multiple_range_change_state)
        self.multiple_text = QLineEdit()
        self.multiple_text.setVisible(False)
        self.multiple_text.setPlaceholderText("Example: [0,1000],[1300,2400],...")
        self.multiple_text.setFixedHeight(28)
        self.multiple_text.editingFinished.connect(self.load_tracking_intervals)

        layout = QHBoxLayout()
        layout.addWidget(self.checkbox)
        layout.addWidget(self.range_slider)
        layout.addWidget(self.multiple_text)
        layout.addWidget(self.multiple_CheckBox)
        self.setLayout(layout)

    def setValue(self, value):
        if value is None:
            self.checkbox.setChecked(False)
            return
        self.checkbox.setChecked(True)
        self.load_tracking_intervals(value)
        if self.checkbox.isChecked():
            self.multiple_CheckBox.setVisible(True)
            multiple = self.multiple_CheckBox.isChecked()
            self.multiple_text.setVisible(multiple)
            self.range_slider.setVisible(not multiple)

    def load_tracking_intervals(self, tracking_intervals=None):
        error_msg = "Please enter a valid interval format"
        n_frames = self.range_slider.maximum()

        try:
            if not tracking_intervals:
                text = self.multiple_text.text().strip()
                if not text:
                    self.multiple_text.clearFocus()
                    self.multiple_CheckBox.setChecked(False)
                    self.emit()
                    return

                tracking_intervals = ast.literal_eval(text)

            if not all(isinstance(item, (list, tuple)) for item in tracking_intervals):
                tracking_intervals = [tracking_intervals]

            assert tracking_intervals
            assert all(tracking_intervals)

            if len(tracking_intervals) == 1:
                self.range_slider.setValue((tracking_intervals[0]))
                self.multiple_text.clearFocus()
                self.multiple_CheckBox.setChecked(False)
                self.emit()
                return
            self.multiple_CheckBox.setChecked(True)

            processed_intervals = []
            for start, end in tracking_intervals:
                start = min(max(int(start), 0), n_frames)
                end = min(max(int(end), 0), n_frames)
                if start > end:
                    start, end = end, start
                if end - start:
                    processed_intervals.append([start, end])

            self.multiple_text.setText(str(processed_intervals)[1:-1])
            self.multiple_text.clearFocus()
            self.emit()
        except (ValueError, SyntaxError, AssertionError, TypeError):
            QMessageBox.warning(self, "Tracking intervals error", error_msg)
            self.multiple_text.setFocus()
            self.multiple_text.setText("")

    def multiple_range_change_state(self, state):
        self.checkbox.setText("Tracking interval" + bool(state) * "s")
        self.range_slider.setVisible(not state)
        self.multiple_text.setVisible(state)

    def checkbox_clicked(self, checked):
        self.multiple_CheckBox.setVisible(checked)
        if checked:
            if self.multiple_CheckBox.isChecked():
                self.multiple_text.setVisible(True)
            else:
                self.range_slider.setVisible(True)
        else:
            self.multiple_text.setVisible(False)
            self.range_slider.setVisible(False)

    def reset(self, n_frames):
        self.range_slider.setRange(0, n_frames)
        self.range_slider.setValue((0, n_frames))
        self.checkbox.setChecked(False)

    def emit(self):
        self.newValue.emit(self.value())

    def value(self) -> list[list[int]] | None:
        if not self.checkbox.isChecked():
            return None
        if self.multiple_CheckBox.isChecked() and self.multiple_text.text():
            return list(ast.literal_eval(self.multiple_text.text()))

        val = list(self.range_slider.value())
        if (
            val[0] == self.range_slider.minimum()
            and val[1] == self.range_slider.maximum()
        ):
            return None
        return [val]



import logging
from argparse import ArgumentParser
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import linear_sum_assignment

from idtrackerai import Video
from idtrackerai.utils import create_dir, wrap_entrypoint

from .matcher import match

plt.rcParams["font.family"] = "STIXgeneral"


def IdMatcherAi(folders: list[Path]):
    logging.info(
        "Matching sessions:\n    "
        + "\n    ".join(map(str, folders[1:]))
        + f"\nwith {folders[0]}"
    )
    master_session = Video.load(folders[0])

    for matching_session in map(Video.load, folders[1:]):
        logging.info("\nMatching %s", matching_session)
        if matching_session.n_animals != master_session.n_animals:
            logging.warning(
                "Different number of animals between\n   "
                f" {matching_session} ({matching_session.n_animals})"
                " and\n   "
                f" {master_session} ({master_session.n_animals})"
            )

        if matching_session.version != master_session.version:
            logging.warning(
                "Different idtracker.ai versions between\n    "
                f"{matching_session} {matching_session.version}"
                " and\n    "
                f"{master_session} {master_session.version}\n"
                "This can cause poor matchings"
            )

        if matching_session.id_image_size != master_session.id_image_size:
            logging.error(
                "Different identification image size between\n    "
                f"{matching_session} {matching_session.id_image_size}"
                " and\n    "
                f"{master_session} {master_session.id_image_size}\n"
                "Check how to define a fixed identification image size in"
                " http://idtracker.ai/en/latest/user_guide/usage.html#identification-image-size"
            )
            continue

        results_path = matching_session.idmatcher_results_path / (
            master_session.session_folder.name
        )
        create_dir(results_path)
        create_dir(results_path / "csv")
        create_dir(results_path / "png")

        direct_matches = match(
            matching_session.id_images_folder, master_session.auto_accumulation_folder
        )
        save_matrix(
            direct_matches,
            results_path,
            "direct_matches",
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )

        indirect_matches = match(
            master_session.id_images_folder, matching_session.auto_accumulation_folder
        ).T
        save_matrix(
            indirect_matches,
            results_path,
            "indirect_matches",
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )

        joined_matches = direct_matches + indirect_matches
        save_matrix(
            joined_matches,
            results_path,
            "joined_matches",
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )

        logging.info(f"{master_session.n_animals} animals in {master_session}")
        logging.info(f"{matching_session.n_animals} animals in {matching_session}")

        if matching_session.n_animals == master_session.n_animals:
            logging.info("Maximizing direct and indirect matches")
            matrix_to_optimize = joined_matches
        elif matching_session.n_animals > master_session.n_animals:
            logging.info("Maximizing indirect matches only")
            matrix_to_optimize = indirect_matches
        elif matching_session.n_animals < master_session.n_animals:
            logging.info("Maximizing direct matches only")
            matrix_to_optimize = direct_matches
        else:
            raise RuntimeError(matching_session.n_animals, master_session.n_animals)

        logging.info("Assigning identities")
        assigned_ids, assignments = linear_sum_assignment(
            matrix_to_optimize, maximize=True
        )

        agreement = (
            joined_matches[assigned_ids, assignments].sum()
            / joined_matches[assigned_ids].sum()
        )

        direct_scores = [
            score_row(direct_matches[assigned_id], assignment)
            for assigned_id, assignment in zip(assigned_ids, assignments)
        ]

        indirect_scores = [
            score_row(indirect_matches[:, assignment], assigned_id)
            for assigned_id, assignment in zip(assigned_ids, assignments)
        ]

        with (results_path / "assignments.csv").open("w", encoding="utf_8") as file:
            file.write("identity, assignment, direct score, indirect score\n")
            for i, assigned_id in enumerate(assigned_ids):
                file.write(
                    f"{assigned_id+1:8d}, {assignments[i]+1:10d},"
                    f" {direct_scores[i]:12.4f}, {indirect_scores[i]:14.4f}\n"
                )

        mean_direct_score = float(np.mean(direct_scores))
        mean_indirect_score = float(np.mean(indirect_scores))
        mean_score = (mean_direct_score + mean_indirect_score) / 2
        save_matrix(
            direct_matches,
            results_path,
            "direct_matches",
            (assignments, assigned_ids, mean_direct_score),
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )
        save_matrix(
            indirect_matches,
            results_path,
            "indirect_matches",
            (assignments, assigned_ids, mean_indirect_score),
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )
        save_matrix(
            joined_matches,
            results_path,
            "joined_matches",
            (assignments, assigned_ids, mean_score),
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )
        logging.info("Results in %s", results_path)
        logging.info(
            "Scores:\n"
            f"    Matching agreement: {agreement:.2%}\n"
            f"    Mean direct score: {mean_direct_score:.2%}\n"
            f"    Mean indirect score: {mean_indirect_score:.2%}\n"
            f"    Total mean score: {mean_score:.2%}"
        )


def score_row(row: np.ndarray, assigned) -> float:
    major_indices = np.argsort(row)[::-1]
    major_value = row[major_indices[0]]
    if major_value == 0:
        return 0
    for major_competitor in major_indices:
        if major_competitor != assigned:
            return float((row[assigned] - row[major_competitor]) / major_value)
    raise ValueError


def path(value: str):
    return_path = Path(value).expanduser().resolve()
    if not return_path.exists():
        raise ValueError()
    return return_path


@wrap_entrypoint
def main():
    for handler in logging.root.handlers:
        handler.setLevel(logging.INFO)

    parser = ArgumentParser()
    parser.add_argument(
        "sessions",
        help="path to the session folder with the results from the first video",
        type=path,
        nargs="+",
    )
    args = parser.parse_args()

    IdMatcherAi(args.sessions)


def save_matrix(
    mat: np.ndarray,
    dir: Path,
    name: str,
    assign: tuple[np.ndarray, np.ndarray, float] | None = None,
    xlabel: str = "",
    ylabel: str = "",
):
    np.savetxt(
        (dir / "csv" / name).with_suffix(".csv"),
        mat,
        "%5d" if mat.shape[1] < 20 else "%d",
        delimiter=",",
    )
    fig, ax = plt.subplots(figsize=(6, 5), dpi=200)
    im = ax.imshow(
        mat,
        interpolation="none",
        extent=(+0.5, mat.shape[1] + 0.5, mat.shape[0] + 0.5, +0.5),
    )

    ax.set(
        title=name.replace("_", " ").capitalize(),
        xlabel=xlabel,
        ylabel=ylabel,
        aspect="auto",
    )
    if assign is not None:
        ax.plot(assign[0] + 1, assign[1] + 1, "rx", ms=8, label="Assignment")
        ax.legend()
        direction = name.split("_")[0]
        ax.set_title(
            ax.get_title() + f" | Assignment {direction} score: {assign[2]:.2%}"
        )

    # show grid
    ax.set_xticks(np.arange(1.5, mat.shape[1]), minor=True)
    ax.set_yticks(np.arange(1.5, mat.shape[0]), minor=True)
    ax.grid(which="minor", color="w")
    ax.tick_params(which="minor", bottom=False, left=False)

    fig.colorbar(im).set_label("Number of matches")
    fig.tight_layout(pad=0.8)
    fig.savefig(str((dir / "png" / name).with_suffix(".png")))



from qtpy.QtCore import QEvent, QSize, Qt, QTimer, Signal  # type: ignore
from qtpy.QtGui import QColor, QFocusEvent, QPainter, QPixmap
from qtpy.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QToolButton,
    QWidget,
)


class CustomList(QListWidget):
    lost_focus = Signal()
    ListChanged = Signal()
    newItemSelected = Signal(object)
    removedItem = Signal(str)

    def __init__(self, max_n_row: int = 5):
        super().__init__()
        self.max_n_row = max_n_row

        self.setAlternatingRowColors(True)

        self.ListChanged.connect(self.update_height)
        self.model().rowsInserted.connect(self.ListChanged.emit)
        self.model().rowsRemoved.connect(self.ListChanged.emit)
        self.model().rowsMoved.connect(self.ListChanged.emit)
        self.itemPressed.connect(self.item_selected)
        self.currentItemChanged.connect(lambda x, y: self.item_selected(x))
        self.selected_item = None

    def focusOutEvent(self, event: QFocusEvent):
        super().focusOutEvent(event)
        if event.reason() is Qt.FocusReason.ActiveWindowFocusReason:
            return  # this fixes drag and drop errors on Wayland
        self.clearSelection()
        item = self.itemWidget(self.selected_item)
        if item:
            item.lost_focus()
        self.newItemSelected.emit(None)
        self.selected_item = None

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if not self.indexAt(event.pos()).isValid():
            self.clearFocus()

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.FontChange:
            QTimer.singleShot(1, self.delayed_update_height)

    def update_height(self):
        # give time to update list items first
        QTimer.singleShot(1, self.delayed_update_height)

    def delayed_update_height(self):
        if self.max_n_row:
            n_rows = max(1, min(5, self.count()))
            item_widget = self.itemWidget(self.item(0))
            row_height = item_widget.height() if item_widget else 25
            self.setFixedHeight(row_height * n_rows + 2 * self.frameWidth())

    def item_selected(self, item: QListWidgetItem | None):
        if self.selected_item == item:
            return
        if self.selected_item is not None:
            widget = self.itemWidget(self.selected_item)
            if isinstance(widget, CustomListItem):
                widget.lost_focus()

        if item is not None:
            self.itemWidget(item).gain_focus()
        self.newItemSelected.emit(item)
        self.selected_item = item

    def remove_item(self):
        item = self.itemAt(self.sender().parent().pos())
        self.item_selected(None)
        self.removedItem.emit(item.data(Qt.ItemDataRole.UserRole))
        self.takeItem(self.row(item))
        self.clearFocus()

    def add_str(self, text: str, color: QColor | int | None = None):
        item = QListWidgetItem()
        cw = CustomListItem(
            text, remove_func=self.remove_item, parent=item, color=color
        )
        item.setData(Qt.ItemDataRole.UserRole, text)
        self.insertItem(-1, item)
        self.setItemWidget(item, cw)

    def getValue(self) -> list[str]:
        return [
            self.item(i).data(Qt.ItemDataRole.UserRole) for i in range(self.count())
        ][::-1]


class CustomListItem(QWidget):
    def __init__(
        self,
        text,
        parent: QListWidgetItem,
        remove_func=None,
        color: None | int | QColor = None,
    ):
        self.list_item = parent
        super().__init__()
        self.selected = False
        self.text = QLabel(text)
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(11, 0, 11, 0)

        if color is not None:
            icon = QLabel()
            icon.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            icon.setFixedSize(10, 10)
            pixmap = QPixmap(icon.size())
            pixmap.fill(Qt.GlobalColor.transparent)
            painter = QPainter(pixmap)
            painter.setBrush(color)
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawEllipse(pixmap.rect())
            painter.end()
            icon.setPixmap(pixmap)
            self.layout().addWidget(icon)

        self.rm_btn = QToolButton()
        self.rm_btn.setText("Remove")
        self.rm_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.rm_btn.clicked.connect(remove_func)
        self.layout().addWidget(self.text)
        self.layout().addWidget(self.rm_btn)
        self.text.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.rm_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.list_item.setSizeHint(QSize(10, self.rm_btn.sizeHint().height() + 4))
        self.update_label_colors()

    def gain_focus(self):
        self.selected = True
        self.text.setStyleSheet(self.selected_stylesheet)

    def lost_focus(self):
        self.selected = False
        self.text.setStyleSheet(self.non_selected_stylesheet)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() in (
            QEvent.Type.PaletteChange,
            QEvent.Type.EnabledChange,
            QEvent.Type.FontChange,
        ):
            self.list_item.setSizeHint(QSize(10, self.rm_btn.sizeHint().height() + 4))
            self.update_label_colors()

    def update_label_colors(self):
        self.selected_stylesheet = (
            "QLabel {color : #"
            + f"{self.palette().highlightedText().color().rgb():x}"
            + f"; font-size:{self.font().pointSize()}pt"
            + "; }"
        )
        self.non_selected_stylesheet = (
            "QLabel {color : #"
            + f"{self.palette().text().color().rgb():x}"
            + f"; font-size:{self.font().pointSize()}pt"
            + "; }"
        )
        if self.selected:
            self.text.setStyleSheet(self.selected_stylesheet)
        else:
            self.text.setStyleSheet(self.non_selected_stylesheet)



import json
import logging
from pathlib import Path
from typing import Iterable

import h5py
import numpy as np

from idtrackerai.network import LearnerClassification, NetworkParams
from idtrackerai.tracker.identity_network import get_predictions_identities


def match(id_images_path: Path, model_path: Path):
    logging.info(
        "Matching images from %s with model from %s", id_images_path, model_path
    )

    id_images_paths = list(id_images_path.glob("id_images_*.hdf5"))

    labels_for_episode = extract_labels_per_episode(id_images_paths)

    set_of_labels = set(np.concatenate(labels_for_episode))
    set_of_labels.discard(0)

    n_img_ids = len(set_of_labels)
    """number of labels in the images to be assigned by the model"""

    model, model_params = load_identification_model(model_path)
    n_model_ids = model_params.n_classes
    """number of classes in the model"""

    matching = np.zeros((n_img_ids, n_model_ids), int)

    for identity in set_of_labels:
        images = extact_images_for_id(id_images_paths, labels_for_episode, identity)
        predictions, softmax_probs = get_predictions_identities(
            model, images, model_params
        )

        matching[identity - 1] = np.bincount(predictions, minlength=n_model_ids + 1)[1:]
    return matching


def extact_images_for_id(
    id_images_file_paths: Iterable[Path],
    labels_per_episode: list[np.ndarray],
    identity: int,
) -> np.ndarray:
    images = []

    for labels, path in zip(labels_per_episode, id_images_file_paths):
        selected_indices = labels == identity
        if not any(selected_indices):
            continue

        with h5py.File(path, "r") as file:
            images.append(
                file["id_images"][selected_indices]  # type: ignore
                if "id_images" in file
                else file["identification_images"][selected_indices]  # type: ignore
            )
    images = np.concatenate(images)
    logging.info("Extracting %d images for identity %d", len(images), identity)
    return images


def extract_labels_per_episode(id_images_file_paths: Iterable[Path]):
    labels = []
    for path in id_images_file_paths:
        with h5py.File(path, "r") as file:
            identities: np.ndarray = file["identities"][:]  # type: ignore

            # v4 compatibility
            if identities.ndim == 2:
                identities = np.squeeze(identities)
            if not issubclass(identities.dtype.type, np.integer):
                identities[np.isnan(identities)] = 0
                identities = identities.astype(int)

            # prerelease compatibility
            identities[identities < 0] = 0

            labels.append(identities)

    return labels


def load_identification_model(model_folder: Path):
    params_path = model_folder / "model_params.json"
    if params_path.is_file():
        with open(params_path, "rb") as file:
            params = json.load(file)
    elif params_path.with_suffix(".npy").is_file():
        params = np.load(params_path.with_suffix(".npy"), allow_pickle=True).item()
    else:
        raise FileNotFoundError(params_path)

    n_classes = (  # 5.1.6 compatibility
        params["n_classes"] if "n_classes" in params else params["number_of_classes"]
    )
    identification_network_params = NetworkParams(
        schedule=params["schedule"],
        n_classes=n_classes,
        architecture="idCNN",
        restore_folder=model_folder,
        model_name=params["model_name"],
        image_size=params["image_size"],
    )

    identification_model = LearnerClassification.load_model(
        identification_network_params
    )
    return identification_model, identification_network_params
